"""Pipeline console script for spatial analysis of pna data.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
from pathlib import Path

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.pna import read
from pixelator.pna.analysis.analysis import (
    ProximityAnalysis,
)
from pixelator.pna.analysis.graph_noise_analysis import KcoreAnalysis, SvdAnalysis
from pixelator.pna.analysis.report import (
    AnalysisSampleReport,
    KCoreReport,
    ProximityReport,
    SvdReport,
)
from pixelator.pna.analysis_engine import AnalysisManager, LoggingSetup
from pixelator.pna.cli.common import output_option
from pixelator.pna.pixeldataset.io import PxlFile

logger = logging.getLogger(__name__)


@click.command(
    "analysis",
    short_help=("Add analysis results to a pxl file."),
    options_metavar="<options>",
)
@click.argument(
    "pxl_file",
    required=True,
    type=click.Path(exists=True),
    metavar="<PIXELFILE>",
)
@click.option(
    "--compute-proximity",
    required=False,
    is_flag=True,
    help="Compute proximity scores matrix",
)
@click.option(
    "--proximity-nbr-of-permutations",
    default=100,
    required=False,
    type=click.IntRange(
        50,
    ),
    show_default=True,
    help="Number of permutations to use when calculating the expected proximity scores",
)
@click.option(
    "--compute-k-cores",
    required=False,
    is_flag=True,
    help="Compute k-core summary tables for each component",
)
@click.option(
    "--compute-svd-var-explained",
    required=False,
    is_flag=True,
    help="Compute variance explained for top 3 singular vectors for each component",
)
@click.option(
    "--svd-nbr-of-pivots",
    default=50,
    required=False,
    type=click.IntRange(10, 200),
    show_default=True,
    help="Number of pivots to use for SVD variance explained computation",
)
@output_option
@click.pass_context
@timer
def analysis(
    ctx,
    pxl_file,
    compute_proximity,
    proximity_nbr_of_permutations,
    compute_k_cores,
    compute_svd_var_explained,
    svd_nbr_of_pivots,
    output,
):
    """Add analysis results to a PXL file."""
    input_files = [pxl_file]
    log_step_start(
        "analysis",
        input_files=input_files,
        output=output,
        compute_proximity=compute_proximity,
        proximity_nbr_of_permutations=proximity_nbr_of_permutations,
        compute_k_cores=compute_k_cores,
        compute_svd_var_explained=compute_svd_var_explained,
        svd_nbr_of_pivots=svd_nbr_of_pivots,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files=input_files, allowed_extensions=("pxl",))

    # sanity check on the analyses
    if not any(
        [
            compute_proximity,
            compute_k_cores,
            compute_svd_var_explained,
        ]
    ):
        logger.warning("All the analysis are disabled, no scores will be computed")

    analysis_to_run = []

    if compute_proximity:
        logger.info("Proximity score computation is activated")
        analysis_to_run.append(
            ProximityAnalysis(n_permutations=proximity_nbr_of_permutations)
        )

    if compute_k_cores:
        logger.info("K-core computation is activated")
        analysis_to_run.append(KcoreAnalysis())

    if compute_svd_var_explained:
        logger.info("SVD variance explained computation is activated")
        analysis_to_run.append(SvdAnalysis(pivots=svd_nbr_of_pivots))

    sample_name = get_sample_name(pxl_file)
    pxl_file = PxlFile(Path(pxl_file))
    analysis_output = create_output_stage_dir(output, "analysis")
    output_file = analysis_output / f"{sample_name}.analysis.pxl"
    metrics = analysis_output / f"{sample_name}.report.json"

    logging_setup = LoggingSetup.from_logger(ctx.obj.get("LOGGER"))
    manager = AnalysisManager(analysis_to_run, logging_setup=logging_setup)
    output_pxl_file_target = PxlFile.copy_pxl_file(pxl_file, output_file)
    pxl_dataset = read(pxl_file.path)
    pxl_dataset_with_analysis = manager.execute(pxl_dataset, output_pxl_file_target)

    write_parameters_file(
        ctx,
        analysis_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-pna analysis",
    )

    proximity_report = ProximityReport() if compute_proximity else None

    k_cores_report = None
    if compute_k_cores:
        k_cores_report = KCoreReport(
            median_average_k_core=pxl_dataset_with_analysis.adata()
            .obs["average_k_core"]
            .median()
        )

    svd_report = None
    if compute_svd_var_explained:
        svd_report = SvdReport(
            median_variance_explained_3d=pxl_dataset_with_analysis.adata()
            .obs.filter(regex="svd_var_expl_s3")
            .median()
        )

    report = AnalysisSampleReport(
        sample_id=sample_name,
        product_id="single-cell-pna",
        proximity=proximity_report,
        k_cores=k_cores_report,
        svd=svd_report,
    )

    report.write_json_file(metrics, indent=4)



"""Base classes for pixelator models.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import json
import os
from pathlib import Path

import pydantic

try:
    from typing import Any, Self
except ImportError:
    from typing_extensions import Self


class SampleReport(pydantic.BaseModel):
    """Base class for all pixelator reports of `single-cell` subcommands.

    :ivar sample_id: The sample id for which the report is generated.
    """

    sample_id: str

    @classmethod
    def from_json(cls, p: Path) -> Self:
        """Initialize an :class:`SampleReport` from a report file.

        :param p: The path to the report file.
        :return: A :class:`SampleReport` object.
        """
        with open(p) as fp:
            json_data = json.load(fp)

        return cls(**json_data)

    def to_json(self, **kwargs: Any) -> str:  # noqa: DOC103
        """Dump the report to a json string.

        :param kwargs: Additional arguments to pass to `json.dumps`.
        :return: The report serialized to JSON as a string.
        """
        return json.dumps(self.model_dump(mode="json"), **kwargs)

    def write_json_file(self, p: str | os.PathLike, **kwargs: Any) -> None:
        """Write a JSON serialized SampleReport to a file.

        Non-existing intermediate directories in the path will be created.

        :param p: The path to the file to write.
        :param kwargs: Additional arguments to pass to pydantics `model_dump_json`.
        """
        Path(p).resolve().parent.mkdir(parents=True, exist_ok=True)

        with open(p, "w") as fp:
            r = self.model_dump_json(**kwargs)
            fp.write(r)



"""Q30 quality statistics for the amplicon regions.

Copyright © 2024 Pixelgen Technologies AB.
"""

import typing
from collections import Counter, defaultdict

import numpy as np
from cutadapt.info import ModificationInfo
from cutadapt.steps import SingleEndStep
from dnaio import SequenceRecord

from pixelator.pna.config import PNAAssay
from pixelator.pna.config.assay import get_position_in_parent
from pixelator.pna.read_processing.statistics import HasCustomStatistics


class QualityStatisticsResult(typing.TypedDict):
    """The result of the quality statistics."""

    fraction_q30_total: float
    fraction_q30_umi1: float
    fraction_q30_pid1: float
    fraction_q30_lbs1: float
    fraction_q30_uei: float
    fraction_q30_lbs2: float
    fraction_q30_pid2: float
    fraction_q30_umi2: float


class QualityStatistics:
    """A pipline step that calculates the Q30 fraction for each region in the amplicon."""

    def __init__(self, data):
        """Initialize the QualityStatistics object.

        :param data: A dictionary with the quality statistics for each region.
        """
        self._region_counters = {}
        for region_id, region_data in data.items():
            self._region_counters[region_id] = Counter(region_data)

    def total_bases(self, region_id: str):
        """Return the total number of bases for a region."""
        return self._region_counters[region_id]["total_bases"]

    def sequences_bases(self, region_id: str):
        """Return the number of sequenced bases for a region."""
        return self._region_counters[region_id]["sequenced_bases"]

    def q30_bases(self, region_id: str):
        """Return the number of bases with quality above 30 for a region."""
        return self._region_counters[region_id]["q30_bases"]

    def get_q30_fraction_total_bases(self, region_id: str):
        """Return the fraction of bases with quality above 30 for a region.

        The fraction is calculated as the number of bases with quality above
        30 divided by the total number bases in the region, even those that were not sequenced.
        """
        r = self._region_counters[region_id]
        if (bases := r["total_bases"]) == 0:
            return 0

        return r["q30_bases"] / bases

    def get_q30_fraction(self, region_id: str):
        """Return the fraction of bases with quality above 30 for a region.

        The fraction is calculated as the number of bases with quality above
        30 divided by the total number of sequenced bases.
        """
        r = self._region_counters[region_id]
        if (bases := r["sequenced_bases"]) == 0:
            return 0

        return r["q30_bases"] / bases

    def __iadd__(self, other):
        """Merge statistics from another object into this one."""
        for region_id, region_data in other._region_counters.items():
            if region_id not in self._region_counters:
                self._region_counters[region_id] = other._region_counters[region_id]
            else:
                self._region_counters[region_id] += other._region_counters[region_id]

        return self

    def collect(self) -> QualityStatisticsResult:
        """Return the quality statistics for the complete amplicon and all regions."""
        return QualityStatisticsResult(
            fraction_q30_total=self.get_q30_fraction("amplicon"),
            fraction_q30_umi1=self.get_q30_fraction("umi-1"),
            fraction_q30_pid1=self.get_q30_fraction("pid-1"),
            fraction_q30_lbs1=self.get_q30_fraction("lbs-1"),
            fraction_q30_uei=self.get_q30_fraction_total_bases("uei"),
            fraction_q30_lbs2=self.get_q30_fraction("lbs-2"),
            fraction_q30_pid2=self.get_q30_fraction("pid-2"),
            fraction_q30_umi2=self.get_q30_fraction("umi-2"),
        )


class QualityProfileStep(SingleEndStep, HasCustomStatistics):
    """A pipeline step that calculates the Q30 fraction for each region in the amplicon."""

    def __init__(self, assay: PNAAssay):
        """Initialize the QualityProfileStep object.

        :param assay: The assay object
        """
        super().__init__()
        self.assay = assay
        self._amplicon_len = assay.get_region_by_id("amplicon").min_len
        self._total_bases = np.zeros(self._amplicon_len, dtype=np.uint64)
        self._total_sequenced_bases = np.zeros(self._amplicon_len, dtype=np.uint64)
        self._q30_bases = np.zeros(self._amplicon_len, dtype=np.uint64)

        self._region_slices = self._get_region_slices()

    def _get_region_slices(self):
        """Return the [start, stop) positions for each region in the amplicon."""
        amplicon_region = self.assay.get_region_by_id("amplicon")
        region_slices = {}

        if amplicon_region is None:
            raise ValueError("Assay does not contain an amplicon region")

        for r in amplicon_region.get_leaves():
            region_slices[r.region_id] = slice(
                *get_position_in_parent(self.assay, r.region_id)
            )

        return region_slices

    def __call__(self, read: SequenceRecord, info: ModificationInfo) -> SequenceRecord:
        """Create a quality profile on each read position for the amplicon."""
        # Create a view into the original quality array
        qual = np.frombuffer(read.qualities_as_bytes(), dtype=np.int8)
        # Convert the vector to integers
        int_qual = qual - 33

        # Increment the total number of bases by one
        # Skip the bases with quality less than 2.
        # Those are mostly LBS region bases that were not sequenced but known from the assay design.
        self._total_bases += 1
        self._total_sequenced_bases += int_qual > 2
        # # Increment the number of bases with quality above 30 by one
        self._q30_bases += int_qual > 30
        return read

    def get_statistics_name(self) -> str:
        """Return the name of the statistics."""
        return "quality_profile"

    def get_statistics(self) -> QualityStatistics:
        """Return the quality statistics."""
        stats: dict[str, dict[str, int]] = defaultdict(dict)

        stats["amplicon"]["q30_bases"] = np.sum(self._q30_bases, dtype=int)
        stats["amplicon"]["sequenced_bases"] = np.sum(
            self._total_sequenced_bases, dtype=int
        )
        stats["amplicon"]["total_bases"] = np.sum(self._total_bases, dtype=int)

        for region_id, region_slice in self._region_slices.items():
            stats[region_id]["q30_bases"] = np.sum(
                self._q30_bases[region_slice], dtype=int
            )
            stats[region_id]["sequenced_bases"] = np.sum(
                self._total_sequenced_bases[region_slice], dtype=int
            )
            stats[region_id]["total_bases"] = np.sum(
                self._total_bases[region_slice], dtype=int
            )

        return QualityStatistics(stats)



"""Copyright © 2023 Pixelgen Technologies AB."""

import logging

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseAdapterQCTestsMixin(BaseWorkflowTestMixin):
    """Base class for adapterqc tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_adapterqc_run"])
    ```
    """

    __stage_key__ = "adapterqc"

    @pytest.mark.dependency(name="test_adapterqc_run", depends=["test_preqc_run"])
    def test_adapterqc_run(self):
        input_files = (self.workdir / "preqc").glob("*.processed.fastq.gz")

        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")
        design = self.__get_data("design")

        command = [
            "--log-file",
            "adapterqc-pixelator.log",
            "single-cell-mpx",
            "adapterqc",
            "--output",
            ".",
            "--design",
            design,
        ]
        if verbose:
            command = ["--verbose"] + command
        if params:
            command += params

        command += input_files
        self.context.run_command("adapterqc", command, input_files)
        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_adapterqc_run"])
    def test_adapterqc_logfile_exist(self):
        assert (self.workdir / "adapterqc-pixelator.log").is_file()

    @pytest.mark.dependency(depends=["test_adapterqc_run"])
    def test_adapterqc_results_folder_exists(self):
        assert (self.workdir / "preqc").is_dir()

    @pytest.mark.dependency(depends=["test_adapterqc_run"])
    def test_adapterqc_processed_output_exists(self):
        processed_files = (self.workdir / "adapterqc").glob("*.processed.fastq.gz")
        for f in processed_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_adapterqc_run"])
    def test_adapterqc_failed_output_exists(self):
        failed_files = (self.workdir / "adapterqc").glob("*.failed.fastq.gz")
        for f in failed_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_adapterqc_run"])
    def test_adapterqc_json_output_exists(self):
        json_files = (self.workdir / "adapterqc").glob("*.report.fastq.gz")
        for f in json_files:
            assert f.is_file()



"""Model for flow of input/output of read counts through processing stages.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import pydantic

from pixelator.mpx.report.models.base import SampleReport


class ReadsDataflowReport(SampleReport):
    """Model for tracking the read counts through all processing stages."""

    input_read_count: int = pydantic.Field(
        ...,
        description="The number of raw input reads from the input fastq files.",
    )

    qc_filtered_read_count: int = pydantic.Field(
        ...,
        description="The number of input reads after basic QC filtering.",
    )

    valid_pbs_read_count: int = pydantic.Field(
        ...,
        description="The number of input reads after QC filtering and with valid PBS1/2 regions.",
    )

    valid_antibody_read_count: int = pydantic.Field(
        ...,
        description=(
            "The number of input reads after QC filtering and with valid PBS1/2 regions and with a valid "
            "antibody barcode."
        ),
    )

    raw_molecule_read_count: int = pydantic.Field(
        ...,
        description=(
            "The number of reads that are attributed to a unique molecule in the sample "
            "after deduplication of reads with close UPIA and UMI sequences. "
            "Note that this should be equal to valid_antibody_read_count."
        ),
    )

    size_filter_fail_molecule_read_count: int = pydantic.Field(
        ...,
        description=(
            "The number of reads in components that do NOT pass "
            "the component size filters."
        ),
    )

    aggregate_molecule_read_count: int | None = pydantic.Field(
        ...,
        description="The number of reads in components identified as aggregates.",
    )

    cell_molecule_read_count: int = pydantic.Field(
        ..., description="The number of reads in cell or aggregate components."
    )

    @pydantic.computed_field(return_type=float)  # type: ignore
    @property
    def fraction_reads_in_molecules(self):
        """Return the fraction of raw input reads in unique molecules."""
        return self.raw_molecule_read_count / self.input_read_count

    @pydantic.computed_field(return_type=float)  # type: ignore
    @property
    def fraction_reads_in_cells(self) -> float:
        """Return the fraction of molecule reads in cells."""
        return self.cell_molecule_read_count / self.input_read_count

    @pydantic.computed_field(return_type=float)  # type: ignore
    @property
    def fraction_valid_pbs_reads(self) -> float:
        """Return the fraction of QC filtered input reads that has a valid PBS region."""
        return self.valid_pbs_read_count / self.input_read_count

    @pydantic.computed_field(return_type=float)  # type: ignore
    @property
    def fraction_valid_antibody_reads(self) -> float:
        """Return the fraction of reads with a valid PBS that have a valid antibody barcode."""
        return self.valid_antibody_read_count / self.input_read_count



"""A registry for shared memory buffers and numpy arrays backed by shared memory.

SharedMemoryRegistry can register objects by name and stores all information needed
to access these objects from different processes.

Copyright © 2025 Pixelgen Technologies AB.
"""

import dataclasses
from multiprocessing.managers import SharedMemoryManager
from multiprocessing.shared_memory import SharedMemory

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt


@dataclasses.dataclass(slots=True, frozen=True, eq=True, init=True)
class ArrayDescriptor:
    """Descriptor for a shared memory array.

    These attributes are used to recreate a numpy array from a shared memory buffer.

    Attributes:
        shape: The shape of the array.
        dtype: The data type of the array.

    """

    shape: tuple[int] | tuple[int, int]
    dtype: npt.DTypeLike


class SharedMemoryRegistry:
    """Helper class that wraps a shared memory manager but registers all allocations by name.

    The registry can be used to allocate shared memory buffers and numpy arrays backed by shared memory.
    The buffers and arrays can be retrieved by name from the registry.
    Only the shared memory buffers are stored in the registry, the numpy arrays are recreated from the
    shared memory buffers when requested.

    This makes recreating the numpy arrays from the SharedMemory objects across processes easier.
    """

    def __init__(self) -> None:
        """Initialize the SharedMemoryHelper."""
        self._manager = SharedMemoryManager()
        self._buffer_registry: dict[str, SharedMemory] = {}
        self._array_registry: dict[str, ArrayDescriptor] = {}

    def __enter__(self) -> Self:
        """Enter the context.

        The SharedMemoryManager will start a management process to handle allocation/deallocation.
        """
        self._manager = self._manager.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Self:
        """Terminate the context."""
        self._manager.__exit__(exc_type, exc_val, exc_tb)
        return self

    def read_only_view(self) -> "ReadOnlySharedMemoryRegistry":
        """Return a read-only view of the registry.

        This view can be shared across threads
        """
        return ReadOnlySharedMemoryRegistry(self)

    def allocate_buffer(self, name: str, n_bytes: int) -> SharedMemory:
        """Allocates a new SharedMemory buffer.

        Args:
            name: The name of the buffer.
            n_bytes: The number of bytes to allocate.

        Returns:
            A SharedMemory object representing the allocated buffer.

        """
        buffer = self._manager.SharedMemory(n_bytes)

        # Release the block from the registry if it already exists
        # Note that the memory will be kept alive until all other processes with
        # references to this object call .unlink() as well or the manager is shutdown.
        if name in self._buffer_registry:
            self._buffer_registry[name].unlink()

        self._buffer_registry[name] = buffer
        return buffer

    def allocate_array(
        self,
        name: str,
        shape: tuple[int] | tuple[int, int],
        dtype: npt.DTypeLike,
        zero_init: bool = True,
    ) -> npt.NDArray:
        """Allocates a new SharedMemory buffer and creates a numpy array backed by this memory.

        Args:
            name: The name to register the array under in the registry.
            shape: The shape of the array.
            dtype: The data type of the array.
            zero_init: Whether to initialize the array with zeros.

        Returns:
            A numpy array backed by shared memory.

        """
        assert 1 <= len(shape) <= 2

        dtype_obj = np.dtype(dtype)
        shm = self.allocate_buffer(name, int(np.prod(shape) * dtype_obj.itemsize))
        array: npt.NDArray = np.ndarray(shape, dtype=dtype_obj, buffer=shm.buf)

        if zero_init:
            if len(shape) == 2:
                array[:, :] = 0
            else:
                array[:] = 0

        self._array_registry[name] = ArrayDescriptor(shape, dtype)
        return array

    def get_buffer(self, name: str) -> SharedMemory | None:
        """Query the registry for a shared memory buffer by name.

        Args:
            name: The name of the buffer

        """
        return self._buffer_registry.get(name)

    def get_array(self, name: str) -> np.ndarray:
        """Query the registry for a shared memory array by name.

        The array will be recreated from the registered shared memory buffer.

        Args:
            name: The name of the array

        """
        desc = self._array_registry.get(name)
        shm = self.get_buffer(name)

        if desc is None or shm is None:
            raise KeyError(f"No array with name '{name}' found in the registry")

        count = np.prod(desc.shape)
        res = np.frombuffer(shm.buf, dtype=desc.dtype, count=count)
        res.shape = desc.shape
        return res

    def unlink_buffer(self, name) -> None:
        """Unlink a shared memory buffer by name.

        Args:
            name: The name of the buffer

        """
        buffer = self._buffer_registry.pop(name, None)
        if buffer is not None:
            buffer.unlink()


class ReadOnlySharedMemoryRegistry:
    """Read-only view of a SharedMemoryRegistry.

    This class allows to create a read-only view of a SharedMemoryRegistry that can be passed to other processes.
    The view can be used to access shared memory buffers and numpy arrays backed by shared memory.
    """

    def __init__(self, registry: SharedMemoryRegistry):
        """Initialize the read-only view.

        Args:
            registry: The SharedMemoryRegistry to create a read-only view of.

        """
        self._buffer_registry = registry._buffer_registry.copy()
        self._array_registry = registry._array_registry.copy()

    def get_buffer(self, name: str) -> SharedMemory | None:
        """Query the registry for a shared memory buffer by name.

        Args:
            name: The name of the buffer

        Returns:
            The shared memory buffer or None if no buffer with the given name is found.

        """
        return self._buffer_registry.get(name)

    def get_array(self, name: str) -> np.ndarray:
        """Query the registry for a shared memory array by name.

        The array will be recreated from the registered shared memory buffer.

        Args:
            name: The name of the array

        Returns:
            The numpy array backed by shared memory.

        """
        desc = self._array_registry.get(name)
        shm = self.get_buffer(name)

        if desc is None or shm is None:
            raise KeyError(f"No array with name '{name}' found in the registry")

        count = np.prod(desc.shape)
        res = np.frombuffer(shm.buf, dtype=desc.dtype, count=count)
        res.shape = desc.shape
        return res



"""Webreport builder.

The WebreportBuilder is used to inject all webreport data into the
template and write the final webreport to a file.

Copyright © 2022 Pixelgen Technologies AB.
"""

import abc
import base64
import dataclasses
import gzip
import json
import logging
from abc import abstractmethod
from pathlib import Path
from typing import (
    IO,
    Any,
    BinaryIO,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import lxml.etree
import semver
from lxml.etree import _Element as LxmlElement
from lxml.html import builder as E

from pixelator.common.report.json_encoder import PixelatorJSONEncoder

from .types import Metrics, QCReportData, SampleInfo

logger = logging.getLogger(__name__)

DEFAULT_QC_REPORT_TEMPLATE = Path(__file__).parent / "template.html"


@dataclasses.dataclass(frozen=True)
class DataInjectionPoint:
    """Data injection point in the template.

    This class is used to represent a data injection point in the template.
    Each injection point is a script tag with a specific data type and dataset id.
    """

    data_type: str
    content_type: str = "application/octet-stream;base64"
    accessor: Callable[[QCReportData], Any] | None = dataclasses.field(default=None)


class BaseQCReportBuilder(metaclass=abc.ABCMeta):
    """Base class for manipulating a QC report from a html template.

    This class is used to inject data into a QC report template to generate a complete report.
    It can also be used to extract data from a report using the `extract_field` method.

    Each custom script tag required by the template is represented by a `DataInjectionPoint`
    object in the `_DATA_FIELDS` class variable.

    Concrete child classes should implement the `write` method to inject the data into the template.
    """

    _DATA_FIELDS: ClassVar[list[DataInjectionPoint]] = []

    def __init__(
        self,
        template: Union[str, Path] = DEFAULT_QC_REPORT_TEMPLATE,
        version_constraints: Optional[List[str]] = None,
        json_options: Optional[Dict[str, Any]] = None,
    ):
        """Construct a QC report builder given a html template.

        :param template: path to the QC report template
        :param version_constraints: list of version constraints for the template
        :param json_options: options to pass to json.dumps when serializing data
        :raises FileNotFoundError: if the template file does not exist
        """
        if not Path(template).exists():
            raise FileNotFoundError(
                f"QC Report template not found at specified location: {template}"
            )
        self.template = Path(template)
        self._json_options = json_options or {"indent": None, "separators": (",", ":")}
        self._data_field_map = {field.data_type: field for field in self._DATA_FIELDS}
        self._version_constraints = version_constraints or []

    def _get_field(self, field_name: str) -> DataInjectionPoint | None:
        return self._data_field_map.get(field_name)

    def _load_template(self) -> Tuple[LxmlElement, LxmlElement]:
        """Load and parse the QC report template.

        :raises AssertionError: if no body tag is found in the template
        :raises AssertionError: if the QC report template version is not supported
        :rtype: Tuple[LxmlElement, LxmlElement]
        """
        logger.debug("Loading QC report template %s", self.template)

        # create a parser
        # Use huge_tree to allow large lxml objects and prevent
        # silent truncation of the html output.
        parser = lxml.etree.HTMLParser(huge_tree=True)

        # load and parse QC report template
        document = None
        with open(self.template, "rb") as f:
            document = lxml.html.parse(f, parser)

        root = document.getroot()
        body = document.find("body")

        if body is None:
            raise AssertionError("Could not find body element in QC report template")

        version = root.xpath('//meta[@name="application-name"]/@data-version')
        version = version[0] if version else ""
        self._check_version_compatibility(version)

        logger.debug("QC report template %s loaded", self.template)
        return root, body

    def _check_version_compatibility(self, version: str):
        if version == "undefined":
            logger.warning("QC report version not found in template")
            return

        for constraint in self._version_constraints:
            check = semver.VersionInfo.parse(version).match(constraint)
            if not check:
                raise AssertionError(
                    f"Unsupported QC report version. Version {version} does "
                    f"not satisfy constraint: {constraint}"
                )

    @abstractmethod
    def write(  # noqa: DOC502
        self,
        fp: BinaryIO,
        sample_info: SampleInfo,
        data: QCReportData,
        metrics_definition_file: Optional[Path] = None,
    ) -> None:
        """Inject given data into the QC report and write the results to a stream.

        :param fp: binary stream to write the report to
        :param sample_info: Sample information
        :param metrics: Metrics for the sample
        :param data: Data for dynamic figures
        :param metrics_definition_file: Path to the metrics definition file
        :raises: AssertionError if not body tag is found in the template

        :example:
             root, body = self._load_template()

            # create the necessary script elements
            elements = []

            if metrics_definition_file is not None:
            # inject the elements in the body (in order for a more structured html)
            for idx, el in enumerate(elements):
                body.insert(idx, el)

            # write to html
            lxml.etree.ElementTree(root).write(fp, method="html", encoding="utf-8")

            logger.debug("Data injected in PNA QC report template %s", self.template)
        """
        raise NotImplementedError

    def _remove_existing_data_element(
        self, body: LxmlElement, field: str | DataInjectionPoint
    ) -> LxmlElement:
        if isinstance(field, str):
            field_desc = self._get_field(field)
            if field_desc is None:
                raise ValueError(f"Field {field} not found in data fields")
        else:
            field_desc = field

        res = body.cssselect(f'script[data-type="{field_desc.data_type}"]')
        if len(res) > 0:
            data_element = res[0]
            parent = data_element.getparent()
            if parent is not None:
                parent.remove(data_element)

        return body

    def _build_elements(self, data: QCReportData):
        elements = {}
        for field in self._DATA_FIELDS:
            if field.accessor is None:
                continue
            else:
                data_field = field.accessor(data)
                if data_field is not None:
                    elements[field.data_type] = self._build_data_element(
                        field.data_type, data_field
                    )

        return elements

    def _build_data_element(self, data_type: str, data: str) -> LxmlElement:
        """Build a script element with the given data and data type."""
        data_descriptor = self._data_field_map[data_type]

        html_elem = E.SCRIPT(
            **{
                "type": data_descriptor.content_type,
                "data-type": data_descriptor.data_type,
                "data-dataset-id": "0",
            }
        )
        html_element_content = data
        if data_descriptor.content_type == "application/octet-stream;base64":
            html_element_content = self._compress_data(data)

        html_elem.text = html_element_content
        return html_elem

    @staticmethod
    def _compress_data(data: str):
        """Compress the data using gzip and encode with base64."""
        return base64.b64encode(gzip.compress(data.encode("utf-8")))

    def extract_field(self, body: LxmlElement, field_name: str) -> str:
        """Extract the data for a given field from the template.

        :param body: the body element of a parsed qc report
        :param field_name: the name of the field to extract
        :returns the data for the given field extracted from the report
        """
        field = self._get_field(field_name)
        if field is None:
            raise ValueError(f"Field {field_name} not found in data fields")

        selector = "script[data-type='{}']".format(field.data_type)
        element = body.cssselect(selector)
        if not element:
            raise ValueError(f"No data found for selector {selector}")

        data = element[0].text or ""
        if field.content_type == "application/octet-stream;base64":
            return gzip.decompress(base64.b64decode(data)).decode("utf-8")

        return data


class PNAQCReportBuilder(BaseQCReportBuilder):
    """Build a qc report from a html template and the required data (CSV and JSON).

    This will parse a qc report template and, using the write method, inject CSV
    and JSON strings into the template using `html <script>` tags under the
    `<body>` tag.

    Tags are labeled with the custom attributes `data-type` and `data-dataset-id`
    to identify the data.

    Note that since we are creating a single sample report (for now) the
    `data-dataset-id` is always set to '0'.

    >>> builder = PNAQCReportBuilder("./path/to/template.html")
    >>> with open("report.html", "wb") as fp:
    ...    builder.write(fp, sample_info, metrics, data)
    """

    VERSIONS_CONSTRAINTS: ClassVar[List[str]] = [">=0.1.0"]

    _DATA_FIELDS = [
        DataInjectionPoint(
            "metric-definitions",
            "application/octet-stream;base64",
        ),
        DataInjectionPoint("metrics", "application/octet-stream;base64"),
        DataInjectionPoint(
            "ranked-component-size",
            "application/octet-stream;base64",
            lambda data: data.ranked_component_size,
        ),
        DataInjectionPoint(
            "component-data",
            "application/octet-stream;base64",
            lambda data: data.component_data,
        ),
        DataInjectionPoint(
            "antibody-counts",
            "application/octet-stream;base64",
            lambda data: data.antibody_counts,
        ),
        DataInjectionPoint(
            "antibody-percentages",
            "application/octet-stream;base64",
            lambda data: data.antibody_percentages,
        ),
        DataInjectionPoint(
            "proximity-data",
            "application/octet-stream;base64",
            lambda data: data.proximity_data,
        ),
    ]

    def __init__(self, template: Union[str, Path] = DEFAULT_QC_REPORT_TEMPLATE):
        """Construct a QC report builder given a html template.

        :param template: path to the QC report template
        :raises FileNotFoundError: if the template file does not exist
        """
        super().__init__(template, self.VERSIONS_CONSTRAINTS)

    def write(  # noqa: DOC502
        self,
        fp: BinaryIO,
        sample_info: SampleInfo,
        data: QCReportData,
        metrics_definitions: Optional[Path | IO[str]] = None,
    ) -> None:
        """Inject given data into the QC report and write the results to a stream.

        :param fp: binary stream to write the report to
        :param sample_info: Sample information
        :param metrics: Metrics for the sample
        :param data: Data for dynamic figures
        :param metrics_definition_file: Path to the metrics definition file
        :raises: AssertionError if not body tag is found in the template
        """
        root, body = self._load_template()

        # create the necessary script elements
        elements = []

        if metrics_definitions is not None:
            elements.append(
                self._build_metric_definition_file_element(metrics_definitions, body)
            )

        elements.append(
            self._build_sample_and_metrics_element(sample_info, data.metrics)
        )

        default_elements = self._build_elements(data)
        drop_elements = {"metrics", "metric-definitions"}
        elements.extend(
            [v for k, v in default_elements.items() if k not in drop_elements]
        )

        # inject the elements in the body (in order for a more structured html)
        for idx, el in enumerate(elements):
            body.insert(idx, el)

        # write to html
        lxml.etree.ElementTree(root).write(fp, method="html", encoding="utf-8")

        logger.debug("Data injected in PNA QC report template %s", self.template)

    def _build_metric_definition_file_element(
        self, metrics_definitions: Path | IO[str], template_body: LxmlElement
    ) -> LxmlElement:
        """Create a lxml HTML object to inject the metrics definition file."""
        # Remove any existing default metric definitions already present in the template
        self._remove_existing_data_element(template_body, "metric-definitions")

        if isinstance(metrics_definitions, Path):
            with open(metrics_definitions, "r") as f:
                data = f.read()
        else:
            data = metrics_definitions.read()

        return self._build_data_element("metric-definitions", data)

    def _build_sample_and_metrics_element(
        self, sample_info: SampleInfo, metrics: Metrics
    ) -> LxmlElement:
        """Create a lxml HTML object to inject the metrics and sample info."""
        metrics_el = E.SCRIPT(
            **{
                "type": "application/octet-stream;base64",
                "data-type": "metrics",
                "data-dataset-id": "0",
            }
        )

        combined = {
            "info": dataclasses.asdict(sample_info),
            "metrics": metrics,
        }
        data = json.dumps(combined, **self._json_options, cls=PixelatorJSONEncoder)
        metrics_el.text = self._compress_data(data)
        return metrics_el



"""Copyright © 2023 Pixelgen Technologies AB."""

import logging
from pathlib import Path

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseAmpliconTestsMixin(BaseWorkflowTestMixin):
    """Base class for amplicon command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_amplicon_run"])
    ```
    """

    __stage_key__ = "amplicon"

    @pytest.mark.dependency(name="test_amplicon_run")
    def test_amplicon_run(self):
        input_files = self.__get_data("input_files")
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")
        design = self.__get_data("design")

        command = [
            "--log-file",
            "amplicon-pixelator.log",
            "single-cell-mpx",
            "amplicon",
            "--output",
            ".",
            "--design",
            str(design),
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += [Path(s).name for s in input_files]

        self.context.run_command("amplicon", command, input_files)

        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_amplicon_run"])
    def test_amplicon_logfile_exist(self):
        assert (self.workdir / "amplicon-pixelator.log").exists()

    @pytest.mark.dependency(depends=["test_amplicon_run"])
    def test_amplicon_results_folder_exists(self):
        assert (self.workdir / "amplicon").exists()

    @pytest.mark.dependency(depends=["test_amplicon_run"])
    def test_amplicon_results_file_exists(self):
        exists_files = (self.workdir / "amplicon").glob("*.merged.fastq.gz")
        for f in exists_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_amplicon_run"])
    def test_amplicon_results_file_not_empty(self):
        not_empty_files = (self.workdir / "amplicon").glob("*.merged.fastq.gz")
        for f in not_empty_files:
            assert f.stat().st_size > 0



"""Copyright © 2023 Pixelgen Technologies AB."""

from pathlib import Path
from typing import Any, Dict

import ruamel.yaml as yaml


class WorkflowConfig:
    """Class used to load and query the workflow tests configuration file."""

    def __init__(self, config_file: Path):
        self.config_file = config_file
        self._config = self._parse(self.config_file)

        self._validate()

    def get_test_config(self, test_name: str) -> Any:
        """Retrieve config data for a specific test case.

        Currently only "small", "single-cell" and "tissue" are implemented.
        You can use arbitrary names for your own tests, and then link your
        :class:`PixelatorWorkfklowTest` subclass to the test name in the config file
        by defining the class variable :attr:`test_id` on the subclass.

        :param test_name: name of the test case as given in the config.
        :returns: the config object for the test case
        """
        return self._config[test_name]

    def keys(self):
        return self._config.keys()

    @classmethod
    def _parse(cls, config_file: Path) -> dict[str, Any]:
        """Load and resolve relative paths in the config file.

        :param config_file: path to the config file
        :return dict[str, Any]: the config object parsed from the `config_file`
        """
        yaml_loader = yaml.YAML(typ="safe")
        with open(str(config_file), "r") as f:
            raw_config = yaml_loader.load(f)

        config = cls._resolve_relative_paths(Path(config_file), raw_config)
        return config

    @classmethod
    def _resolve_relative_paths(cls, config_file: Path, config: Dict[str, Any]):
        """Resolve relative paths in the config file.

        All relative paths in the config file are resolved relative to the parent
        directory of the config file.

        :param config_file: path to the config file
        :param config: the config object parsed from the `config_file`
        :returns: the config object with relative paths resolved
        """
        for test_id, test_config in config.items():
            panel_file = test_config.get("panel_file")
            if panel_file:
                test_config["panel_file"] = (config_file.parent / panel_file).resolve()

            input_files = test_config.get("input_files")
            if input_files:
                new_input_files = [
                    (config_file.parent / f).resolve() for f in input_files
                ]
                test_config["input_files"] = new_input_files

        return config

    def _validate(self):
        cfg = self._config
        for wf_name, wf in cfg.items():
            if wf.get("panel_file") and wf.get("panel"):
                raise ValueError(
                    f"Both `panel` and `panel_file` are defined for "
                    f"workflow {wf_name}. Only one can be set."
                )



"""Console script for pixelator (collapse).

Copyright © 2022 Pixelgen Technologies AB.
"""

import json
import logging
from pathlib import Path

import click
import polars as pl

from pixelator.common.utils import (
    create_output_stage_dir,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.pna.cli.common import (
    design_option,
    output_option,
    panel_option,
    threads_option,
)
from pixelator.pna.collapse.independent.collapser import (
    RegionCollapser,
    SingleUMICollapseSampleReport,
)
from pixelator.pna.collapse.paired.collapser import MoleculeCollapser
from pixelator.pna.collapse.utilities import split_collapse_inputs
from pixelator.pna.config import load_antibody_panel, pna_config
from pixelator.pna.utils import get_demux_filename_info

logger = logging.getLogger("collapse")


def validate_mismatches(ctx, param, value):
    """Validate the --mismatches commandline option.

    :param ctx: The click context
    :param param: The click parameter name
    :param value: The click value
    :returns: The validated value
    """
    try:
        value = int(value)
        if value >= 1:
            return value
    except ValueError:
        pass

    try:
        value = float(value)
        if 0 <= value < 1:
            return value
    except ValueError:
        pass

    raise click.BadParameter("Must be an integer >= 1 or a float in range [0, 1)")


@click.command(
    "collapse",
    short_help=(
        "Detect duplicates and perform error correction on demultiplexed PNA data (parquet)."
    ),
    options_metavar="<options>",
)
@click.argument(
    "input_files",
    nargs=-1,
    required=True,
    type=click.Path(exists=True),
    metavar="PARQUET_FILES",
)
@click.option(
    "--mismatches",
    default=2,
    required=False,
    type=click.UNPROCESSED,
    callback=validate_mismatches,
    show_default=True,
    help=("The number of mismatches allowed when collapsing."),
)
@click.option(
    "--algorithm",
    required=False,
    type=click.Choice(["directional", "cluster"]),
    default="directional",
    show_default=True,
    help=("The network based algorithm to use for collapsing."),
)
@threads_option
@design_option
@panel_option
@output_option
@click.pass_context
@timer
def collapse(
    ctx,
    input_files,
    design,
    panel,
    output,
    mismatches,
    algorithm,
    threads,
):
    """Collapse Molecular Pixelation data (FASTQ) to remove duplicates and perform error correction."""  # noqa
    # log input parameters
    log_step_start(
        "collapse",
        panel=panel,
        design=design,
        output=output,
        mismatches=mismatches,
        algorithm=algorithm,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files=input_files, allowed_extensions=("parquet",))

    assay = pna_config.get_assay(design)
    panel = load_antibody_panel(pna_config, panel)

    # create the output directory
    collapse_output = create_output_stage_dir(output, "collapse")

    checked_parquet = split_collapse_inputs(input_files)
    if isinstance(checked_parquet, tuple):
        logger.info("Detected independent UMI1 and UMI2 demuxed input.")

        umi1_files, umi2_files = checked_parquet
        return process_independent_files(
            umi1_files,
            umi2_files,
            assay=assay,
            panel=panel,
            collapse_output=collapse_output,
            mismatches=mismatches,
            algorithm=algorithm,
            threads=threads,
            ctx=ctx,
        )

    logger.info("Detected paired UMI1 and UMI2 demuxed input.")
    return process_paired_input(
        ctx, input_files, panel, mismatches, algorithm, threads, assay, collapse_output
    )


def process_paired_input(
    ctx, input_files, panel, mismatches, algorithm, threads, assay, collapse_output
) -> None:
    """Process paired UMI1 and UMI2 demuxed files.

    This will run the collapse process on the files together

    Args:
        ctx: The click context.
        input_files: The UM1 and UMI2 demuxed files.
        panel: The panel configuration.
        mismatches: The number of mismatches allowed when error-correcting.
        algorithm: The collapse strategy to use.
        threads: The number of threads to use.
        assay: The assay configuration.
        collapse_output: The output filename for the collapsed data.

    """
    if len(input_files) == 1:
        logger.info("Detected single input file.")
        input_file = input_files[0]
        sample_name, part_number = get_demux_filename_info(input_file)

        if part_number:
            logger.info(
                "Found part number in input, will use that part number in output files."
            )
            output_file = (
                collapse_output
                / f"{sample_name}.part_{part_number:03d}.collapsed.parquet"
            )
            metrics_output = (
                collapse_output / f"{sample_name}.part_{part_number:03d}.report.json"
            )
            param_file = (
                collapse_output / f"{sample_name}.part_{part_number:03d}.meta.json"
            )
        else:
            output_file = collapse_output / f"{sample_name}.collapsed.parquet"
            metrics_output = collapse_output / f"{sample_name}.report.json"
            param_file = collapse_output / f"{sample_name}.meta.json"
    else:
        logger.info("Detected multiple input files.")
        sample_name, part_number = get_demux_filename_info(input_files[0])

        output_file = collapse_output / f"{sample_name}.collapsed.parquet"
        metrics_output = collapse_output / f"{sample_name}.report.json"
        param_file = collapse_output / f"{sample_name}.meta.json"

    write_parameters_file(
        ctx,
        param_file,
        command_path="pixelator single-cell-pna collapse",
    )

    collapser = MoleculeCollapser(
        assay=assay,
        panel=panel,
        output=output_file,
        max_mismatches=mismatches,
        algorithm=algorithm,
        threads=threads,
    )

    with collapser as c:
        for f in input_files:
            c.process_file(Path(f))

    stats = collapser.statistics()
    stats.add_summary_statistics(pl.scan_parquet(output_file))

    encoder = json.JSONEncoder(indent=4)
    with open(metrics_output, "w") as f:
        f.write(encoder.encode(stats.to_dict()))


def process_independent_files(
    umi1_files,
    umi2_files,
    *,
    assay,
    panel,
    collapse_output,
    mismatches,
    algorithm,
    threads=-1,
    ctx=None,
):
    """Process independent UMI1 and UMI2 demuxed files.

    This will run the collapse process on all UMI1 files and UMI2 files successively.

    Args:
        umi1_files: The UMI1 demuxed files.
        umi2_files: The UMI2 demuxed files.
        assay: The assay configuration.
        panel: The panel configuration.
        collapse_output: The output filename for the collapsed data.
        mismatches: The number of mismatches allowed when error-correcting.
        algorithm: The collapse strategy to use.
        threads: The number of threads to use.
        ctx: The click context.

    Returns:
        A tuple with umi1 and umi2 output file paths

    """
    umi1_outputs = []
    umi2_outputs = []

    min_parallel_chunk_size = 500

    umi1_collapser = RegionCollapser(
        assay=assay,
        panel=panel,
        region_id="umi-1",
        threads=threads,
        algorithm=algorithm,
        min_parallel_chunk_size=min_parallel_chunk_size,
        max_mismatches=mismatches,
    )

    with umi1_collapser as c:
        for umi1_file in umi1_files:
            res = get_demux_filename_info(umi1_file)

            if not res:
                raise click.ClickException(
                    f"Could not extract sample name and part number from file name: {umi1_file}"
                )

            sample_name, part_number = res
            base_output_name = f"{sample_name}.collapse.m1.part_{part_number:03d}"
            output_file = collapse_output / f"{base_output_name}.parquet"
            metrics_output = collapse_output / f"{base_output_name}.report.json"

            # Allow skipping writing the parameters file when not running from a click context
            if ctx:
                write_parameters_file(
                    ctx,
                    collapse_output / f"{base_output_name}.meta.json",
                    command_path="pixelator single-cell-pna collapse",
                )

            c.process_file(umi1_file, output=output_file)
            umi1_outputs.append(output_file)

            stats = umi1_collapser.statistics()
            report = SingleUMICollapseSampleReport(
                sample_id=sample_name, product_id="single-cell-pna", **stats.to_dict()
            )
            report.write_json_file(metrics_output, indent=4)

    del umi1_collapser

    umi2_collapser = RegionCollapser(
        assay=assay,
        panel=panel,
        region_id="umi-2",
        threads=threads,
        algorithm=algorithm,
        min_parallel_chunk_size=min_parallel_chunk_size,
        max_mismatches=mismatches,
    )

    with umi2_collapser as c:
        for umi2_file in umi2_files:
            res = get_demux_filename_info(umi2_file)

            if not res:
                raise click.ClickException(
                    f"Could not extract sample name and part number from file name: {umi2_file}"
                )

            sample_name, part_number = res
            base_output_name = f"{sample_name}.collapse.m2.part_{part_number:03d}"
            output_file = collapse_output / f"{base_output_name}.parquet"
            metrics_output = collapse_output / f"{base_output_name}.report.json"

            # Allow skipping writing the parameters file when not running from a click context
            if ctx:
                write_parameters_file(
                    ctx,
                    collapse_output / f"{base_output_name}.meta.json",
                    command_path="pixelator single-cell-pna collapse",
                )

            c.process_file(umi2_file, output=output_file)
            umi2_outputs.append(output_file)

            stats = umi2_collapser.statistics().to_dict()
            report = SingleUMICollapseSampleReport(
                sample_id=sample_name, product_id="single-cell-pna", **stats
            )
            report.write_json_file(metrics_output, indent=4)

    del umi2_collapser

    return (umi1_outputs, umi2_outputs)



"""Console script for pixelator pna amplicon.

Copyright © 2024 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
import sys
from pathlib import Path

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.pna.amplicon import amplicon_fastq
from pixelator.pna.amplicon.report import AmpliconSampleReport
from pixelator.pna.cli.common import (
    design_option,
    logger,
    output_option,
    threads_option,
)
from pixelator.pna.config import pna_config
from pixelator.pna.utils import get_read_sample_name, is_read_file


@click.command(
    "amplicon",
    short_help=("process diverse raw pixel data (FASTQ) formats into common amplicon"),
    options_metavar="<options>",
)
@click.argument(
    "fastq_1",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ_1",
)
@click.argument(
    "fastq_2",
    nargs=1,
    required=False,
    type=click.Path(exists=True),
    metavar="FASTQ_2",
)
@output_option
@design_option
@click.option(
    "--sample-name",
    default=None,
    show_default=False,
    type=click.STRING,
    help=(
        "Override the basename of the output fastq file. "
        "Default is the basename of the first input file "
        "without extension and read 1 identifier."
    ),
)
@click.option(
    "--mismatches",
    default=0.1,
    required=False,
    type=click.FloatRange(0.0, 0.9),
    show_default=True,
    help="The number of mismatches allowed (given as fraction)",
)
@click.option(
    "--remove-polyg",
    required=False,
    is_flag=True,
    help="Remove PolyG sequences (length of 10 or more)",
)
@click.option(
    "--quality-cutoff",
    required=False,
    default=20,
    help="Remove bases from the tail with a Phred score lower then the cutoff value.",
)
@click.option(
    "--skip-input-checks",
    default=False,
    is_flag=True,
    type=click.BOOL,
    help="Skip all check on the filename of input fastq files.",
)
@click.option(
    "--force-run",
    default=False,
    is_flag=True,
    type=click.BOOL,
    help=(
        "If more than 50% of reads are filtered, this indicates some serious problem with the input data and the amplicon command"
        "will fail with an exit status of 1. Setting this flag will force an exit status of 0."
    ),
)
@threads_option
@click.pass_context
@timer
def amplicon(
    ctx,
    fastq_1: Path,
    fastq_2: Path | None,
    sample_name: str | None,
    output: str,
    design: str,
    mismatches: int,
    remove_polyg: bool,
    quality_cutoff: int,
    skip_input_checks: bool,
    force_run: bool,
    threads: int,
):
    """Process diverse raw pixel data (FASTQ) formats into common amplicon."""
    # log input parameters

    error_level = logging.WARNING if skip_input_checks else logging.ERROR

    log_step_start(
        "amplicon",
        fastq1=fastq_1,
        fastq_2=fastq_2,
        output=output,
        design=design,
        remove_polyg=remove_polyg,
        quality_cutoff=quality_cutoff,
        skip_input_checks=skip_input_checks,
        force_run=force_run,
    )

    fastq_inputs = [Path(fastq_1)]
    if fastq_2:
        fastq_inputs.append(Path(fastq_2))
    # some basic sanity check on the input files
    sanity_check_inputs(
        fastq_inputs,
        allowed_extensions=("fastq.gz", "fq.gz", "fastq", "fq", "fastq.zst", "fq.zst"),
    )

    # create output folder if it does not exist
    amplicon_output = create_output_stage_dir(output, "amplicon")
    r1_sample_name = get_read_sample_name(fastq_1)

    # Some checks on the input files
    # - check if there are read 1 and read2 identifiers in the filename
    # - check if the sample name is the same for read1 and read2
    # no need to check this if only one fastq file is given.
    if fastq_2 is not None:
        if not is_read_file(fastq_1, "r1"):
            msg = "Read 1 file does not contain a recognised read 1 suffix."
            logger.log(level=error_level, msg=msg)
            if not skip_input_checks:
                ctx.exit(1)

        if fastq_2 and not is_read_file(fastq_2, "r2"):
            msg = "Read 2 file does not contain a recognised read 2 suffix."
            logger.log(level=error_level, msg=msg)
            if not skip_input_checks:
                ctx.exit(1)

        r2_sample_name = get_read_sample_name(fastq_2) if fastq_2 else None

        if r1_sample_name != r2_sample_name:
            msg = (
                f"The sample name for read1 and read2 is different:\n"
                f'"{r1_sample_name}" vs "{r2_sample_name}"\n'
                "Did you pass the correct files?"
            )
            logger.log(level=error_level, msg=msg)
            if not skip_input_checks:
                ctx.exit(1)

    sample_name = sample_name or r1_sample_name
    output_file = amplicon_output / f"{sample_name}.amplicon.fq.zst"
    json_file = amplicon_output / f"{sample_name}.report.json"

    write_parameters_file(
        ctx,
        amplicon_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-pna amplicon",
    )

    msg = f"Creating amplicon for {','.join(str(p) for p in fastq_inputs)}"
    logger.info(msg)

    assay = pna_config.get_assay(design)
    if assay is None:
        msg = f"Could not find assay design with name '{design}'"
        logger.error(msg)
        return ctx.exit(1)

    stats = amplicon_fastq(
        inputs=fastq_inputs,
        assay=assay,
        output=output_file,
        mismatches=mismatches,
        poly_g_trimming=remove_polyg,
        quality_cutoff=quality_cutoff,
        threads=threads,
    )

    report = AmpliconSampleReport(
        sample_id=sample_name, product_id="single-cell-pna", **stats.as_dict()
    )

    report.write_json_file(json_file, indent=4)

    if report.fraction_discarded_reads > 0.5:
        logger.error(
            "The number of reads in the output file is less than half of the input file. "
            "This may indicate a problem with the amplicon design."
        )
        if not force_run:
            ctx.exit(1)



"""Copyright © 2025 Pixelgen Technologies AB."""

import pandas as pd

from pixelator.pna.pixeldataset.legacy import PNALegacyPixelDataset


def polarity_score(pxl_dataset: PNALegacyPixelDataset) -> pd.DataFrame:
    """Extract polarity scores from a PNAPixelDataset."""
    return pxl_dataset._backend._datastore.read_polarization()


def colocalization_score(pxl_dataset: PNALegacyPixelDataset) -> pd.DataFrame:
    """Extract colocalization scores from a PNAPixelDataset."""
    return pxl_dataset._backend._datastore.read_colocalization()



"""Module for aggregating pixeldatasets.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
from typing import (
    List,
)

import pandas as pd
import polars as pl
from anndata import AnnData
from anndata import concat as concatenate_anndata

from pixelator.mpx.pixeldataset import PixelDataset
from pixelator.mpx.pixeldataset.precomputed_layouts import aggregate_precomputed_layouts
from pixelator.mpx.pixeldataset.utils import (
    _enforce_edgelist_types,
    update_metrics_anndata,
)

logger = logging.getLogger(__name__)


def _get_attr_and_index_by_component(attribute, sample_names, datasets):
    for name, dataset in zip(sample_names, datasets):
        attr = getattr(dataset, attribute, None)
        if attr is not None:
            attr["sample"] = pd.Categorical([name] * len(attr))
            attr["component"] = pd.Categorical(
                attr["component"].astype(str) + "_" + attr["sample"].astype(str)
            )
            attr.set_index(["component"])
            yield attr


def _concatenate_edgelists(datasets, sample_names):
    with pl.StringCache():
        concatenated = datasets[0].edgelist_lazy
        concatenated = concatenated.with_columns(
            sample=pl.lit(sample_names[0], dtype=pl.Categorical)
        )
        concatenated = concatenated.collect()

        for idx, subsequent in enumerate(datasets[1:], start=1):
            concatenated = concatenated.extend(
                subsequent.edgelist_lazy.collect().with_columns(
                    sample=pl.lit(sample_names[idx], dtype=pl.Categorical)
                )
            )

        concatenated = concatenated.with_columns(
            component=pl.concat_str(
                pl.col("component"), pl.col("sample"), separator="_"
            )
        )

    return concatenated


def simple_aggregate(
    sample_names: List[str],
    datasets: List[PixelDataset],
    ignore_edgelists: bool = False,
) -> PixelDataset:
    """Aggregate samples in a simple way (see caveats).

    Aggregating samples in a simplistic fashion. This function should only
    be used if the dataset you merge have been generated with the same panel.

    It will concatenate all dataframes in the underlying PixelDataset instances,
    and add a new column called sample. New indexes will be formed from from the
    `sample` and `component` columns.

    The metadata dictionary will contain one key per sample.

    Since the edgelists will take up considerable amounts of memory it is possible to
    ignore them when aggregating the data. This will mean that they are not directly
    available for downstream analysis.

    :param sample_names: an iterable of the sample names to use for each dataset
    :param datasets: an iterable of the datasets you want to aggregate
    :param ignore_edgelists: ignoring merging the edgelists, leaving them empty in the
                             resulting PixelDataset. Defaults to False.

    :raises AssertionError: If not all pre-conditions are meet.
    :return: a PixelDataset instance with all the merged samples
    :rtype: PixelDataset
    """
    if not (len(datasets)) > 1:
        raise AssertionError(
            "There must be two or more datasets and names passed to `aggregate`"
        )
    if not len(sample_names) == len(datasets):
        raise AssertionError(
            "There must be as many sample names provided as there are dataset"
        )
    if not len(set(sample_names)) == len(sample_names):
        raise AssertionError("All provided sample names must be unique")

    all_var_identical = all(
        map(
            lambda x: x.adata.var.index.equals(datasets[0].adata.var.index),
            datasets,
        )
    )
    if not all_var_identical:
        raise AssertionError("All datasets must have identical `vars`")

    def _add_sample_name_as_obs_col(adata, name):
        adata.obs["sample"] = name
        return adata

    tmp_adatas = concatenate_anndata(
        {
            name: _add_sample_name_as_obs_col(dataset.adata, name)
            for name, dataset in zip(sample_names, datasets)
        },
        axis=0,
        index_unique="_",
    )
    adata = AnnData(tmp_adatas.X, obs=tmp_adatas.obs, var=datasets[0].adata.var)
    update_metrics_anndata(adata=adata, inplace=True)

    if ignore_edgelists:
        edgelists = pd.DataFrame()
    else:
        edgelists = _enforce_edgelist_types(
            _concatenate_edgelists(datasets, sample_names).to_pandas()
        )

    polarizations = pd.concat(
        _get_attr_and_index_by_component(
            "polarization", datasets=datasets, sample_names=sample_names
        ),
        axis=0,
    )
    colocalizations = pd.concat(
        _get_attr_and_index_by_component(
            "colocalization", datasets=datasets, sample_names=sample_names
        ),
        axis=0,
    )
    metadata = {
        "samples": {
            name: dataset.metadata for name, dataset in zip(sample_names, datasets)
        }
    }

    precomputed_layouts = aggregate_precomputed_layouts(
        [(name, dataset) for name, dataset in zip(sample_names, datasets)],
        all_markers=set(datasets[0].adata.var.index),
    )

    return PixelDataset.from_data(
        adata=adata,
        edgelist=edgelists,
        polarization=polarizations,
        colocalization=colocalizations,
        metadata=metadata,
        precomputed_layouts=precomputed_layouts,
        copy=False,
        allow_empty_edgelist=ignore_edgelists,
    )



"""Module for PixelDataset data stores, e.g. pxl files and similar.

Copyright © 2024 Pixelgen Technologies AB.
"""

from __future__ import annotations

import io
import json
import logging
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Literal,
    Optional,
    Protocol,
    Type,
    runtime_checkable,
)

import anndata as ad
import h5py
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from anndata import AnnData
from fsspec.implementations.zip import ZipFileSystem

from pixelator.common.exceptions import PixelatorBaseException
from pixelator.common.utils import batched
from pixelator.mpx.pixeldataset.precomputed_layouts import PreComputedLayouts

if TYPE_CHECKING:
    from pixelator.mpx.pixeldataset import PixelDataset

from pixelator.common.types import PathType

logger = logging.getLogger(__name__)


class PixelDataStoreError(PixelatorBaseException):
    """Base class for all PixelDataStore related exceptions."""

    pass


class CannotOverwriteError(PixelDataStoreError):
    """Raised when a file is attempted to be overwritten."""

    pass


class FileFormatNotRecognizedError(PixelDataStoreError):
    """Raised when the file format of a .pxl file is not recognized."""

    pass


class EdgelistNotFoundError(PixelDataStoreError):
    """Raised when the edgelist is not found in a .pxl file."""

    pass


class CannotGuessPixelDataStoreError(PixelDataStoreError):
    """Raised when the pixel data store format cannot be guessed from the path."""

    pass


@runtime_checkable
class PixelDataStore(Protocol):
    """Interface for pixel data storage.

    The methods here should be implemented by any pixel data store.

    """

    @staticmethod
    def from_path(path: PathType) -> PixelDataStore:
        """Get a PixelDataStore from the provided path.

        :param path: The path to the pixel data store.
        :return: A pixel data store.
        :rtype: PixelDataStore
        :raises CannotGuessPixelDatastoreError: If the datastore format cannot be guessed from the path.
        """
        return PixelDataStore.guess_datastore_from_path(path)

    @staticmethod
    def guess_datastore_from_path(path: PathType) -> PixelDataStore:
        """Guess the pixel data store format based on the given path.

        :param path: The path to the pixel data store.
        :return: The guessed pixel data store format.
        :rtype: PixelDataStore
        :raises CannotGuessPixelDatastoreError: If the datastore format cannot be guessed from the path.
        """
        if str(path).endswith(".pxl"):
            return ZipBasedPixelFile.guess_file_format(path)
        raise CannotGuessPixelDataStoreError(
            f"Could not guess datastore from path {path}"
        )

    def file_format_version(self) -> Optional[int]:
        """Return the file format version of the pixel data store."""
        ...

    def read_anndata(self) -> AnnData:
        """Read the pixel data as an AnnData object.

        :return: The pixel data as an AnnData object.
        :rtype: AnnData
        """
        ...

    def write_anndata(self, anndata: AnnData) -> None:
        """Write the given AnnData object to the pixel data store.

        :param anndata: The AnnData object to write.
        """
        ...

    def read_metadata(self) -> Dict[str, Any]:
        """Read the metadata associated with the pixel data store.

        :return: The metadata as a dictionary.
        :rtype: Dict[str, Any]
        """
        ...

    def write_metadata(self, metadata: Dict[str, Any]) -> None:
        """Write the given metadata to the pixel data store.

        :param metadata: The metadata to write.
        """
        ...

    def read_dataframe(self, key: str) -> pd.DataFrame:
        """Read a dataframe from the pixel data store.

        :param key: The key of the dataframe to read.
        :return: The dataframe.
        :rtype: pd.DataFrame
        """
        ...

    def read_dataframe_lazy(self, key: str) -> Optional[pl.LazyFrame]:
        """Read a lazy dataframe from the pixel data store.

        :param key: The key of the dataframe to read.
        :return: The lazy dataframe.
        :rtype: Optional[pl.LazyFrame]
        """
        ...

    def write_dataframe(
        self,
        dataframe: pd.DataFrame,
        key: str,
        partitioning: Optional[list[str]] = None,
    ) -> None:
        """Write the given dataframe to the pixel data store.

        If a partitioning is provided, the dataframe will be partitioned accordingly.
        This may not be supported by all data stores. If not supported, this
        option will be ignored.

        :param dataframe: The dataframe to write.
        :param key: The key to write the dataframe to.
        :param partitioning: The (optional) partitioning to use when
                             writing the dataframe.
        """
        ...

    def read_edgelist(self) -> pd.DataFrame:
        """Read an edgelist from the pixel data store.

        :return: The edgelist.
        :rtype: pd.DataFrame
        """
        ...

    def read_edgelist_lazy(self) -> pl.LazyFrame:
        """Read a lazy edgelist from the pixel data store.

        :return: The lazy edgelist.
        :rtype: pl.LazyFrame
        """
        ...

    def read_polarization(self) -> Optional[pd.DataFrame]:
        """Read the polarization data from the pixel data store.

        :return: The polarization data, or None if it doesn't exist.
        :rtype: Optional[pd.DataFrame]
        """
        ...

    def read_colocalization(self) -> Optional[pd.DataFrame]:
        """Read the colocalization data from the pixel data store.

        :return: The colocalization data, or None if it doesn't exist.
        :rtype: Optional[pd.DataFrame]
        """
        ...

    def write_edgelist(
        self, edgelist: pd.DataFrame, partitioning: Optional[list[str]] = None
    ) -> None:
        """Write the given edgelist to the pixel data store.

        If a partitioning is provided, the edgelist will be partitioned accordingly.
        This may not be supported by all data stores. If not supported, this
        option will be ignored.

        :param edgelist: The edgelist to write.
        :param partitioning: The (optional) partitioning to use when
                             writing the dataframe.
        """
        ...

    def write_polarization(self, polarization: pd.DataFrame) -> None:
        """Write the given polarization data to the pixel data store.

        :param polarization: The polarization data to write.
        """
        ...

    def write_colocalization(self, colocalization: pd.DataFrame) -> None:
        """Write the given colocalization data to the pixel data store.

        :param colocalization: The colocalization data to write.
        """
        ...

    def save(self, dataset: PixelDataset, force_overwrite: bool = False) -> None:
        """Save the given PixelDataset to the pixel data store.

        :param dataset: The PixelDataset to save.
        :param force_overwrite: Whether to force overwrite an existing file.
        """
        ...

    def read_precomputed_layouts(
        self,
    ) -> PreComputedLayouts:
        """Read pre-computed layouts from the data store."""
        ...

    def write_precomputed_layouts(
        self,
        layouts: PreComputedLayouts,
    ) -> None:
        """Write pre-computed layouts to the data store.

        :param layouts: The pre-computed layouts to write.
        :param collapse_to_single_dataframe: Whether to collapse the layouts into
                                             a single dataframe before writing.
        """
        ...


class _CustomZipFileSystem(ZipFileSystem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def find(self, path, maxdepth=None, withdirs=False, detail=False, **kwargs):
        if maxdepth is not None and maxdepth < 1:
            raise ValueError("maxdepth must be at least 1")

        result = {}

        def _below_max_recursion_depth(path):
            if not maxdepth:
                return True

            depth = len(path.split("/"))
            return depth <= maxdepth

        for zip_info in self.zip.infolist():
            file_name = zip_info.filename
            if not file_name.startswith(path.lstrip("/")):
                continue

            # zip files can contain explicit or implicit directories
            # hence the need to either add them directly or infer them
            # from the file paths
            if zip_info.is_dir():
                if withdirs:
                    if not result.get(file_name) and _below_max_recursion_depth(
                        file_name
                    ):
                        result[file_name.strip("/")] = (
                            self.info(file_name) if detail else None
                        )
                    continue
                else:
                    continue  # Skip along to the next entry if we don't want to add the dirs

            if not result.get(file_name):
                if _below_max_recursion_depth(file_name):
                    result[file_name] = self.info(file_name) if detail else None

                # Here we handle the case of implicitly adding the
                # directories if they have been requested
                if withdirs:
                    directories = file_name.split("/")
                    for i in range(1, len(directories)):
                        dir_path = "/".join(directories[:i]).strip(
                            "/"
                        )  # remove the trailing slash, as this is not expected
                        if not result.get(dir_path) and _below_max_recursion_depth(
                            dir_path
                        ):
                            result[dir_path] = {
                                "name": dir_path,
                                "size": 0,
                                "type": "directory",
                            }

        return result if detail else sorted(list(result.keys()))


class ZipBasedPixelFile(PixelDataStore):
    """Superclass for all zip-based pixel data stores."""

    FILE_FORMAT_VERSION: int = 1

    FILE_FORMAT_VERSION_KEY: str = "file_format_version"
    ANNDATA_KEY: str
    COLOCALIZATION_KEY: str
    EDGELIST_KEY: str
    METADATA_KEY: str
    LAYOUTS_KEY: str = "layouts.parquet"
    POLARIZATION_KEY: str

    def __init__(self, path: PathType) -> None:
        """Create a zip-based pixel data store."""
        self.path = path
        self._file_system_handle = None
        self._current_mode: Literal["r", "w", "a"] | None = None

    def __enter__(self):
        """Enter the context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        self.close()

    def __del__(self) -> None:
        """Ensure file handle is closed when object is deleted."""
        # Make sure the file handles are closed when the object is deleted
        self.close()

    def _setup_file_system(self, mode):
        files_system = _CustomZipFileSystem(fo=self.path, mode=mode, allowZip64=True)

        # For now we are overwriting the zip open method to force it to
        # always have force_zip=True, otherwise it won't work for large file
        # This is related to this issue in fsspec:
        # https://github.com/fsspec/filesystem_spec/issues/1474
        def custom_open(f):
            return partial(f, force_zip64=True)

        files_system.zip.open = custom_open(files_system.zip.open)
        self._file_system_handle = files_system
        self._current_mode = mode

    @property
    def _file_system(self):
        if not self._file_system_handle:
            self._setup_file_system("r")
        return self._file_system_handle

    def _set_to_write_mode(self):
        if not (self._current_mode == "w" or self._current_mode == "a"):
            if self._file_system_handle:
                self._file_system_handle.close()
            if Path(self.path).is_file():
                self._setup_file_system("a")
            else:
                self._setup_file_system("w")

    def _set_to_read_mode(self):
        if self._current_mode != "r":
            if self._file_system_handle:
                self._file_system_handle.close()
            self._setup_file_system("r")

    def _check_if_writeable(self, key):
        try:
            if self._file_system.exists(key):
                raise CannotOverwriteError(
                    f"Data with key {key} already exists in pixel file. We currently do not "
                    "support overwriting. Please save your .pxl file under a new name."
                )
        except FileNotFoundError:
            return True

    def close(self):
        """Close the file system handle."""
        if self._file_system_handle:
            self._file_system_handle.close()

    @staticmethod
    def from_file(path) -> PixelDataStore:
        """Guess the file format of the given path and returns the a PixelDataStore."""
        return ZipBasedPixelFile.guess_file_format(path)

    @staticmethod
    def guess_file_format(path: PathType) -> PixelDataStore:
        """Guess the file format of the given path and returns the a PixelDataStore."""
        file_system = ZipFileSystem(fo=path, mode="r")
        members = [file_["name"] for file_ in file_system.ls("/")]
        file_format: Optional[Type[ZipBasedPixelFile]] = None
        if ZipBasedPixelFileWithParquet.EDGELIST_KEY in members:
            file_format = ZipBasedPixelFileWithParquet
        if ZipBasedPixelFileWithCSV.EDGELIST_KEY in members:
            file_format = ZipBasedPixelFileWithCSV
        if not file_format:
            raise FileFormatNotRecognizedError(
                f"Could not identify file format of input file {path}"
            )
        file_system.close()
        return file_format(path)

    def file_format_version(self) -> Optional[int]:
        """Return the file format version of the .pxl file."""
        return self.read_metadata().get(self.FILE_FORMAT_VERSION_KEY)

    def read_anndata(self) -> AnnData:
        """Read the AnnData object from the .pxl file."""
        self._set_to_read_mode()
        with self._file_system.open(self.ANNDATA_KEY, "rb") as af:
            with h5py.File(af, "r") as f:
                data = ad.experimental.read_elem(f["/"])
        return data

    def write_anndata(self, anndata: AnnData) -> None:
        """Write the given AnnData object to the .pxl file."""
        self._set_to_write_mode()
        self._check_if_writeable(self.ANNDATA_KEY)
        with self._file_system.open(self.ANNDATA_KEY, "wb", compression=None) as af:
            bio = io.BytesIO()
            with h5py.File(bio, "w") as f:
                ad.experimental.write_elem(f, "/", anndata)
            af.write(bio.getbuffer())

    def read_edgelist(self) -> pd.DataFrame:
        """Read the edgelist from the .pxl file."""
        df = self.read_dataframe(self.EDGELIST_KEY)
        if df is None:
            raise EdgelistNotFoundError("Edgelist not found in pxl file")
        return df

    def read_edgelist_lazy(self) -> pl.LazyFrame:
        """Read the edgelist lazily from the .pxl file."""
        lazy_data_frame = self.read_dataframe_lazy(self.EDGELIST_KEY)
        if lazy_data_frame is None:
            raise EdgelistNotFoundError("Edgelist not found in pxl file")
        return lazy_data_frame

    def read_polarization(self) -> Optional[pd.DataFrame]:
        """Read the polarization data from the .pxl file."""
        return self.read_dataframe(self.POLARIZATION_KEY)

    def read_colocalization(self) -> Optional[pd.DataFrame]:
        """Read the colocalization data from the .pxl file."""
        return self.read_dataframe(self.COLOCALIZATION_KEY)

    def read_metadata(self) -> Dict:
        """Read the metadata from the .pxl file."""
        self._set_to_read_mode()
        try:
            with self._file_system.open(self.METADATA_KEY, "r") as f:
                return json.loads(f.read())
        except FileNotFoundError:
            return {}

    def read_precomputed_layouts(
        self,
    ) -> PreComputedLayouts:
        """Read pre-computed layouts from the .pxl file."""
        layouts_lazy = self.read_dataframe_lazy(self.LAYOUTS_KEY)
        if layouts_lazy is None:
            return PreComputedLayouts.create_empty()
        return PreComputedLayouts(layouts_lazy=layouts_lazy)

    def write_metadata(self, metadata: Dict[str, Any]) -> None:
        """Write the given metadata to the .pxl file."""
        self._set_to_write_mode()
        self._check_if_writeable(self.METADATA_KEY)
        with self._file_system.open(self.METADATA_KEY, "w") as f:
            f.write(json.dumps(metadata))

    def write_edgelist(
        self, edgelist: pd.DataFrame, partitioning: Optional[list[str]] = None
    ) -> None:
        """Write the given edgelist to the .pxl file."""
        self.write_dataframe(edgelist, self.EDGELIST_KEY, partitioning)

    def write_polarization(self, polarization: pd.DataFrame) -> None:
        """Write the given polarization data to the .pxl file."""
        self.write_dataframe(polarization, self.POLARIZATION_KEY)

    def write_colocalization(self, colocalization: pd.DataFrame) -> None:
        """Write the given colocalization data to the .pxl file."""
        self.write_dataframe(colocalization, self.COLOCALIZATION_KEY)

    def write_precomputed_layouts(
        self,
        layouts: Optional[PreComputedLayouts],
    ) -> None:
        """Write pre-computed layouts to the data store."""
        if layouts is None:
            logger.debug("No layouts to write, will skip.")
            return

        logger.debug("Starting to write layouts...")

        self._set_to_write_mode()
        self._check_if_writeable(self.LAYOUTS_KEY)

        # This is a work around for the fact that sinking into parquet files
        # from multiple sources is not supported. We therefore do this somewhat
        # round about thing of first writing the parquet files to
        # a temporary directory and then zipping them into the .pxl file.
        with TemporaryDirectory(prefix="pixelator-") as tmp_dir:
            tmp_path = Path(tmp_dir)
            local_tmp_target = tmp_path / "local.layouts.parquet"
            layouts.write_parquet(local_tmp_target, partitioning=layouts.partitioning)

            for file_ in local_tmp_target.rglob("*"):
                if file_.is_file():
                    # TODO Make sure written without compression
                    file_name = file_.relative_to(local_tmp_target)
                    self._file_system.zip.write(
                        file_, arcname=f"{self.LAYOUTS_KEY}/{file_name}"
                    )

        logger.debug("Completed writing layouts...")

    def save(self, dataset: PixelDataset, force_overwrite: bool = False) -> None:
        """Save the given PixelDataset to the .pxl file."""
        path = Path(self.path)
        if path.exists():
            if force_overwrite:
                logger.warning("Overwriting existing .pxl file at %s", self.path)
                path.unlink()
            else:
                raise CannotOverwriteError(
                    "Cannot overwrite existing .pxl file at %s, use `force_overwrite=True` "
                    "overwrite the existing file." % self.path
                )

        self._set_to_write_mode()
        logger.debug("Writing anndata")
        self.write_anndata(dataset.adata)
        # TODO Consider using default partitioning for edgelist here
        logger.debug("Writing edgelist")
        self.write_edgelist(dataset.edgelist)
        logger.debug("Writing metadata")
        self.write_metadata(dataset.metadata)
        if dataset.polarization is not None and dataset.polarization.shape[0] > 0:
            logger.debug("Writing polarization scores")
            self.write_polarization(dataset.polarization)

        if dataset.colocalization is not None and dataset.colocalization.shape[0] > 0:
            logger.debug("Writing colocalization scores")
            self.write_colocalization(dataset.colocalization)

        if not dataset.precomputed_layouts.is_empty:
            logger.debug("Writing precomputed layouts")
            # This speeds things up massively when you have many, very small
            # layouts, like we do in some test data.
            self.write_precomputed_layouts(
                dataset.precomputed_layouts,
            )

        logger.debug("PixelDataset saved to %s", self.path)


class ZipBasedPixelFileWithCSV(ZipBasedPixelFile):
    """A zip-based pixel data store that uses csv files for data storage.

    This is only provided for backward compatibility, the ZipBasedPixelFileWithParquet
    should be used instead of this for all new code.
    """

    ANNDATA_KEY: str = "adata.h5ad"
    EDGELIST_KEY: str = "edgelist.csv.gz"
    METADATA_KEY: str = "metadata.json"
    POLARIZATION_KEY: str = "polarization.csv.gz"
    COLOCALIZATION_KEY: str = "colocalization.csv.gz"

    def __init__(self, path: PathType) -> None:
        """Create a zip-based pixel file using csv files to store dataframes."""
        super().__init__(path)

    def write_dataframe(
        self,
        dataframe: pd.DataFrame,
        key: str,
        partitioning: Optional[list[str]] = None,
    ) -> None:
        """Write the given dataframe to the .pxl file."""
        # Note that partitioning will be ignored here
        self._set_to_write_mode()

        self._check_if_writeable(key)

        with self._file_system.open(key, "wb", compression=None) as f:
            dataframe.to_csv(f, compression="gzip", index=False)

    def read_dataframe(self, key: str) -> Optional[pd.DataFrame]:
        """Read a dataframe from the .pxl file."""
        self._set_to_read_mode()
        return self._read_dataframe_from_zip(key)

    def read_dataframe_lazy(self, key: str) -> Optional[pl.LazyFrame]:
        """Read a dataframe lazily from a zip file (NB: Not implemented!)."""
        raise NotImplementedError(
            "You are trying to read data lazily from a csv-based pxl file. "
            "This is currently not supported. "
            "You can fix this issue by converting your pxl file by saving it "
            "as a parquet based pxl file."
        )

    def write_precomputed_layouts(
        self,
        layouts: Optional[PreComputedLayouts],
    ) -> None:
        """Write pre-computed layouts to the data store (NB: Not implemented!)."""
        raise NotImplementedError(
            "You are trying to write precomputed layouts to a csv-based pxl file. "
            "This is not supported. Please save your pxl file as a parquet based pxl file "
            "instead."
        )

    def _read_dataframe_from_zip(self, key: str) -> Optional[pd.DataFrame]:
        self._set_to_read_mode()
        try:
            with self._file_system.open(key, "rb") as f:
                return pl.read_csv(f).to_pandas()
        except KeyError:
            return None


class ZipBasedPixelFileWithParquet(ZipBasedPixelFile):
    """A zip-based pixel data store that uses parquet files for data storage."""

    ANNDATA_KEY: str = "adata.h5ad"
    EDGELIST_KEY: str = "edgelist.parquet"
    METADATA_KEY: str = "metadata.json"
    POLARIZATION_KEY: str = "polarization.parquet"
    COLOCALIZATION_KEY: str = "colocalization.parquet"

    def __init__(self, path) -> None:
        """Create a zip-based pixel file using parquet files to store dataframes."""
        super().__init__(path)

    def write_dataframe(
        self,
        dataframe: pd.DataFrame,
        key: str,
        partitioning: Optional[list[str]] = None,
    ) -> None:
        """Write the given dataframe to the .pxl file.

        Optionally provided a `partitioning` to create a hive partitioned parquet file,
        i.e. a directory structure with one level per partitioning provided, and parquet
        files as leaves.

        :param dataframe: The dataframe to write.
        :param key: The key of the dataframe to write.
        :param partitioning: The partitioning to use when writing the dataframe.
        """
        DEFAULT_COMPRESSION = "zstd"

        self._set_to_write_mode()
        if partitioning:
            file_options = ds.ParquetFileFormat().make_write_options(
                compression=DEFAULT_COMPRESSION
            )
            for _, data in dataframe.groupby(partitioning, observed=True):
                ds.write_dataset(
                    pa.Table.from_pandas(data, preserve_index=False),
                    f"{key}/",
                    filesystem=self._file_system,
                    format="parquet",
                    partitioning_flavor="hive",
                    partitioning=partitioning,
                    use_threads=False,
                    file_options=file_options,
                    existing_data_behavior="overwrite_or_ignore",
                )
            return

        self._check_if_writeable(key)
        pq.write_table(
            pa.Table.from_pandas(dataframe, preserve_index=False),
            where=key,
            filesystem=self._file_system,
            compression=DEFAULT_COMPRESSION,
            # We want all the data to go into one
            # parquet row group to allow for maximum compression
            # when we write the data here.
            # the `or 1` ensures that it does not raise if
            # the dataframe is empty.
            row_group_size=len(dataframe) or 1,
        )

    def read_dataframe(self, key: str) -> Optional[pd.DataFrame]:
        """Read a dataframe from the .pxl file."""
        self._set_to_read_mode()
        return self._read_dataframe_from_zip(key)

    def read_dataframe_lazy(self, key: str) -> Optional[pl.LazyFrame]:
        """Read a dataframe lazily from a zip file."""
        self._set_to_read_mode()
        return self._read_dataframe_from_zip_lazy(key)

    def _read_dataframe_from_zip(self, key: str) -> Optional[pd.DataFrame]:
        self._set_to_read_mode()
        df = self._read_dataframe_from_zip_lazy(key)
        if df is None:
            return None
        return df.collect().to_pandas()

    def _read_dataframe_from_zip_lazy(self, key: str) -> Optional[pl.LazyFrame]:
        try:
            self._set_to_read_mode()
            dataset = ds.dataset(
                key,
                filesystem=self._file_system,
                partitioning="hive",
                format="parquet",
                partition_base_dir=key,
            )
            return pl.scan_pyarrow_dataset(dataset)
        except FileNotFoundError:
            return None



"""Functions for filtering and annotating of pixel data in edge list format.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
import typing
from pathlib import Path
from typing import Literal, Optional

import numba
import numpy as np
import pandas as pd
import polars as pl
from anndata import AnnData
from graspologic_native import leiden

from pixelator import __version__
from pixelator.common.annotate import filter_components_sizes
from pixelator.common.annotate.aggregates import call_aggregates
from pixelator.common.annotate.cell_calling import find_component_size_limits
from pixelator.common.annotate.constants import (
    MINIMUM_NBR_OF_CELLS_FOR_ANNOTATION,
)
from pixelator.common.config import AntibodyPanel
from pixelator.common.exceptions import PixelatorBaseException
from pixelator.common.report.models import SummaryStatistics
from pixelator.mpx.graph.utils import components_metrics, edgelist_metrics
from pixelator.mpx.pixeldataset import SIZE_DEFINITION, PixelDataset
from pixelator.mpx.pixeldataset.utils import edgelist_to_anndata
from pixelator.mpx.report.models.annotate import AnnotateSampleReport

# TODO
# Work around for issue with numba and multithreading
# I got the solution from here:
# https://stackoverflow.com/questions/68131348/training-a-python-umap-model-hangs-in-a-multiprocessing-process
numba.config.THREADING_LAYER = "omp"

logger = logging.getLogger(__name__)


class NoCellsFoundException(PixelatorBaseException):
    """Raised when no cells are found in the edge list."""

    pass


def annotate_components(
    input: str,
    panel: AntibodyPanel,
    output: str,
    output_prefix: str,
    metrics_file: str,
    min_size: Optional[int],
    max_size: Optional[int],
    dynamic_filter: Optional[Literal["both", "min", "max"]],
    aggregate_calling: bool,
    verbose: bool,
) -> None:
    """Create a pixeldataset from a provided edgelist.

    This function takes as input an `edge list` dataframe (csv) that has been
    generated with `pixelator cluster`.

    The function filters the components in the edge list by component size
    (`min_size` and `max_size`) which is defined as the number of edges.
    Note that the large components (above `max_size`) are not filtered out
    but are marked as `is_filtered`.

    When `dynamic_filter` is "min" the `min_size` will be overruled by using a rank
    based approach, if it is "max" the `max_size` will be overruled, and if it is
    "both" both `min_size` and `max_size` will be overruled. This argument can be
    None to disable the dynamic filtering.

    The raw component metrics with the `is_filtered` column are stored in a csv file
    before filtering the edge list. This allows debugging and the generation of
    figures in the report.

    The filtered edge list is then converted to an `AnnData` object. The values used
    for filtering the components are stored in the `uns` layer.

    When `aggregate_calling` is True aggregates will be called based on marker
    specificity. This will add to keys to the AnnDAta object's `obs`,
    `is_aggregate` and `tau`. The former gives an estimate of if the component
    is an aggregate or not, and the latter contains the computed tau specificity
    score.

    The following files and QC figures are generated after the filtering:

    - a dataframe with the components metrics before filtering (csv)
    - a PixelDataset with the filtered AnnData and edge list (zip)

    :param input: the path to the edge list dataframe (parquet)
    :param panel: the AntibodyPanel of the panel used to generate the data
    :param output: the path to the output folder
    :param output_prefix: the prefix to prepend to the files (sample name)
    :param metrics_file: the path to a JSON file to write metrics
    :param min_size: the minimum size a component must have
    :param max_size: the maximum size a component must have
    :param dynamic_filter: use a rank based approach to define the min and or max size
    :param aggregate_calling: activate aggregate calling
    :param verbose: run if verbose mode when true
    :returns: None
    :rtype: None
    :raises: RuntimeError if max_size is smaller than min_size
    """
    logger.debug("Parsing edge list %s", input)

    # load data (edge list in data frame format)
    edgelist = pl.read_parquet(input).to_pandas()

    # get component metrics from the edge list
    component_metrics = components_metrics(edgelist=edgelist)

    # get the distribution of component sizes
    component_sizes = component_metrics[SIZE_DEFINITION].to_numpy()

    # obtain min/max size bounds using a dynamic rank based approach
    if dynamic_filter in ["min", "both"]:
        min_size = find_component_size_limits(
            component_sizes=component_sizes, direction="lower"
        )
    if dynamic_filter in ["max", "both"]:
        max_size = find_component_size_limits(
            component_sizes=component_sizes, direction="upper"
        )

    # check that mask size is not smaller than min size
    if max_size is not None and min_size is not None and max_size <= min_size:
        raise RuntimeError(
            f"max_size={max_size} must be greater than min_size={min_size}"
        )

    # filter the components by size, the function does no filter
    # the component sizes, it returns a numpy array evaluates to True
    # if the component pass the filters
    is_filtered_arr = filter_components_sizes(
        component_sizes=component_sizes,
        min_size=min_size,
        max_size=max_size,
    )
    component_metrics["is_filtered"] = is_filtered_arr

    # Calculate some metrics on the components that fail the size filter
    input_cell_count = component_metrics.shape[0]
    input_read_count = component_metrics.loc[:, "reads"].sum()
    size_filter_fail_mask = ~is_filtered_arr
    size_filter_fail_cell_count = np.sum(size_filter_fail_mask, dtype=int)
    size_filter_fail_molecule_count = component_metrics.loc[
        size_filter_fail_mask, "molecules"
    ].sum()
    size_filter_fail_read_count = component_metrics.loc[
        size_filter_fail_mask, "reads"
    ].sum()

    # save the components metrics (raw)
    component_info_file_path = (
        Path(output) / f"{output_prefix}.raw_components_metrics.csv.gz"
    )
    component_metrics.to_csv(
        component_info_file_path,
        header=True,
        index=True,
        sep=",",
        compression="gzip",
    )

    # filter the components metrics using the is_filtered flag
    filtered_component_metrics = component_metrics[is_filtered_arr]

    # filter the edge list
    filtered_edgelist = edgelist[
        edgelist["component"].isin(filtered_component_metrics.index)
    ]

    # convert the filtered edge list to AnnData
    adata = edgelist_to_anndata(edgelist=filtered_edgelist, panel=panel)

    # components clustering
    if adata.n_obs > MINIMUM_NBR_OF_CELLS_FOR_ANNOTATION:
        # perform the unsupervised clustering
        cluster_components(adata=adata, inplace=True)
    else:
        logger.warning(
            ("Skipping clustering since there are less than %s components"),
            MINIMUM_NBR_OF_CELLS_FOR_ANNOTATION,
        )

    if aggregate_calling:
        call_aggregates(adata, inplace=True)

    # create filtered PixelDataset and save it
    adata.uns["version"] = __version__
    metadata = {"version": __version__, "sample": output_prefix}
    dataset = PixelDataset.from_data(
        edgelist=filtered_edgelist, adata=adata, metadata=metadata
    )
    dataset.save(
        Path(output) / f"{output_prefix}.annotate.dataset.pxl", force_overwrite=True
    )

    edgelist_metrics_dict = edgelist_metrics(filtered_edgelist)
    adata_metrics_dict = anndata_metrics(adata)

    molecules_per_a_pixel_per_cell_stats = SummaryStatistics.from_series(
        filtered_component_metrics["mean_molecules_per_a_pixel"]
    )
    a_pixel_count_per_b_pixel_per_cell_stats = SummaryStatistics.from_series(
        filtered_component_metrics["mean_a_pixels_per_b_pixel"]
    )
    b_pixel_count_per_a_pixel_per_cell_stats = SummaryStatistics.from_series(
        filtered_component_metrics["mean_b_pixels_per_a_pixel"]
    )

    report = AnnotateSampleReport(
        sample_id=output_prefix,
        marker_count=edgelist_metrics_dict["marker_count"],
        a_pixel_count=edgelist_metrics_dict["a_pixel_count"],
        b_pixel_count=edgelist_metrics_dict["b_pixel_count"],
        molecule_count_per_a_pixel_per_cell_stats=molecules_per_a_pixel_per_cell_stats,
        a_pixel_count_per_b_pixel_per_cell_stats=a_pixel_count_per_b_pixel_per_cell_stats,
        b_pixel_count_per_a_pixel_per_cell_stats=b_pixel_count_per_a_pixel_per_cell_stats,
        fraction_molecules_in_largest_component=edgelist_metrics_dict[
            "fraction_molecules_in_largest_component"
        ],
        fraction_pixels_in_largest_component=edgelist_metrics_dict[
            "fraction_pixels_in_largest_component"
        ],
        **adata_metrics_dict,
        input_cell_count=input_cell_count,
        input_read_count=input_read_count,
        size_filter_fail_cell_count=size_filter_fail_cell_count,
        size_filter_fail_molecule_count=size_filter_fail_molecule_count,
        size_filter_fail_read_count=size_filter_fail_read_count,
    )

    # save metrics (JSON)
    report.write_json_file(Path(metrics_file), indent=4)


def _cluster_components_using_leiden(
    adata: AnnData, resolution: float = 1.0, random_seed: Optional[int] = None
) -> None:
    """Carry out a leiden clustering on the components."""
    # It should be ok to run this over all vs all even on a dense matrix
    # since it shouldn't apply to more than a few thousande components.
    connections = adata.obsp["connectivities"]
    edgelist = list(
        (adata.obs.index[i], adata.obs.index[j], connections[i, j])
        for i, j in zip(*connections.nonzero())
    )
    _, partitions = leiden(
        edgelist,
        resolution=resolution,
        seed=random_seed,
        use_modularity=True,
        # These parameters are used to sync up the native implementation with
        # the python implementation we originally used.
        iterations=1,
        randomness=0.001,
        trials=1,
        starting_communities=None,
    )
    partitions_df = pd.DataFrame.from_dict(partitions, orient="index")
    adata.obs["leiden"] = partitions_df
    adata.obs = adata.obs.astype({"leiden": "category"})


def cluster_components(
    adata: AnnData,
    obsmkey: Optional[Literal["clr", "log1p"]] = "clr",
    inplace: bool = True,
    random_seed: Optional[int] = None,
) -> Optional[AnnData]:
    """Cluster the components based on their antibody counts.

    Clusters components based on their clr transformed antibody counts using
    the k-nearest neighbors, UMAP and leiden algorithms.

    It requires that the `obsmkey` is  present in the `obsm`
    layer of the input `adata` object.

    A new column called `leiden` will be added to `obs` containing the
    cluster ids.

    A new column called `X_umap` will be added to `obsm` containing the
    coordinates of the UMAP manifold.

    :param adata: AnnData object to do the clustering on
    :param obsmkey: Key to access the values `obsm` layer of `adata`
    :param inplace: If `True` performs the operation inplace on `adata`
    :param random_seed: If set this seed will be used to seed the random number
                        generators used when calculating neighbors, building the umap
                        and for the leiden clustering.
    :returns: a new Anndata object if `inplace` is `True` or None
    :rtype: Optional[AnnData]
    :raises: AssertionError if `obsmkey` is missing
    """
    # Import here as it is a slow import
    import scanpy as sc

    if obsmkey not in adata.obsm:
        raise AssertionError(f"Input AnnData is missing '{obsmkey}'")

    logger.debug("Computing clustering for %i components", adata.n_obs)

    if not inplace:
        adata = adata.copy()

    # perform the clustering (using default values)
    sc.pp.neighbors(
        adata,
        use_rep=obsmkey,
        n_neighbors=15,
        random_state=random_seed if random_seed else 0,
    )
    sc.tl.umap(adata, min_dist=0.5, random_state=random_seed if random_seed else 0)
    _cluster_components_using_leiden(
        adata, resolution=1.0, random_seed=random_seed if random_seed else None
    )

    logger.debug("Clustering computed %i clusters", adata.obs["leiden"].nunique())
    return None if inplace else adata


class AnnotateAnndataStatistics(typing.TypedDict):
    """Typed dict for the statistics of an AnnData object."""

    cell_count: int

    total_marker_count: int

    molecule_count: int

    read_count: int

    molecule_count_per_cell_stats: SummaryStatistics
    read_count_per_cell_stats: SummaryStatistics
    a_pixel_count_per_cell_stats: SummaryStatistics
    b_pixel_count_per_cell_stats: SummaryStatistics
    marker_count_per_cell_stats: SummaryStatistics
    a_pixel_b_pixel_ratio_per_cell_stats: SummaryStatistics

    aggregate_count: Optional[int]
    reads_in_aggregates_count: Optional[int]
    molecules_in_aggregates_count: Optional[int]

    min_size_threshold: Optional[int]
    max_size_threshold: Optional[int]
    doublet_size_threshold: Optional[int]
    fraction_potential_doublets: Optional[float]
    n_edges_to_split_potential_doublets: Optional[int]


def anndata_metrics(adata: AnnData) -> AnnotateAnndataStatistics:
    """Collect metrics from an AnnData object.

    :param adata: the AnnData object
    :returns: a dictionary of different metrics
    """
    molecule_count = adata.obs["molecules"].sum()
    read_count = adata.obs["reads"].sum()
    molecules_per_cell_stats = SummaryStatistics.from_series(adata.obs["molecules"])
    reads_per_cell_stats = SummaryStatistics.from_series(adata.obs["reads"])
    a_pixels_per_cell_stats = SummaryStatistics.from_series(adata.obs["a_pixels"])
    b_pixels_per_cell_stats = SummaryStatistics.from_series(adata.obs["b_pixels"])
    markers_per_cell_stats = SummaryStatistics.from_series(adata.obs["antibodies"])
    a_pixel_b_pixel_ratio_per_cell_stats = SummaryStatistics.from_series(
        adata.obs["a_pixel_b_pixel_ratio"]
    )

    metrics: AnnotateAnndataStatistics = {
        "cell_count": adata.n_obs,
        "total_marker_count": adata.n_vars,
        "molecule_count": molecule_count,
        "read_count": read_count,
        "molecule_count_per_cell_stats": molecules_per_cell_stats,
        "read_count_per_cell_stats": reads_per_cell_stats,
        "a_pixel_count_per_cell_stats": a_pixels_per_cell_stats,
        "b_pixel_count_per_cell_stats": b_pixels_per_cell_stats,
        "marker_count_per_cell_stats": markers_per_cell_stats,
        "a_pixel_b_pixel_ratio_per_cell_stats": a_pixel_b_pixel_ratio_per_cell_stats,
        "aggregate_count": None,
        "reads_in_aggregates_count": None,
        "molecules_in_aggregates_count": None,
        "min_size_threshold": None,
        "max_size_threshold": None,
        "doublet_size_threshold": None,
        "fraction_potential_doublets": None,
        "n_edges_to_split_potential_doublets": None,
    }

    # Tau type will only be available if it has been added in the annotate step
    if "tau_type" in adata.obs:
        aggregates_mask = adata.obs["tau_type"] != "normal"
        number_of_aggregates = np.sum(aggregates_mask)
        metrics["aggregate_count"] = number_of_aggregates
        metrics["reads_in_aggregates_count"] = adata[aggregates_mask].obs["reads"].sum()
        metrics["molecules_in_aggregates_count"] = (
            adata[aggregates_mask].obs["molecules"].sum()
        )

    if "min_size_threshold" in adata.uns:
        metrics["min_size_threshold"] = adata.uns["min_size_threshold"]

    if "max_size_threshold" in adata.uns:
        metrics["max_size_threshold"] = adata.uns["max_size_threshold"]

    if "doublet_size_threshold" in adata.uns:
        metrics["doublet_size_threshold"] = adata.uns["doublet_size_threshold"]

    if "is_potential_doublet" in adata.obs:
        metrics["fraction_potential_doublets"] = adata.obs[
            "is_potential_doublet"
        ].mean()

    if "n_edges_to_split_doublet" in adata.obs:
        metrics["n_edges_to_split_potential_doublets"] = adata.obs[
            "n_edges_to_split_doublet"
        ].sum()

    return metrics



"""Console script for pixelator (preqc).

Copyright © 2022 Pixelgen Technologies AB.
"""

from shutil import which

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_extension,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.cli.common import (
    design_option,
    logger,
    output_option,
)
from pixelator.mpx.config import config
from pixelator.mpx.qc import qc_fastq


@click.command(
    "preqc",
    short_help=(
        "process raw pixel data (FASTQ) to perform QC,"
        " filtering, trimming and remove duplicates"
    ),
    options_metavar="<options>",
)
@click.argument(
    "fastq_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ",
)
@click.option(
    "--trim-front",
    default=0,
    required=False,
    type=click.INT,
    show_default=True,
    help="Trim N bases from the front of the reads",
)
@click.option(
    "--trim-tail",
    default=0,
    required=False,
    type=click.INT,
    show_default=True,
    help="Trim N bases from the tail of the reads",
)
@click.option(
    "--max-length",
    default=None,
    required=False,
    type=click.INT,
    show_default=False,
    help=(
        "The maximum length (bases) of a read (longer reads will be trimmed off).\n"
        "If you set this argument it will overrule the value from the chosen design"
    ),
)
@click.option(
    "--min-length",
    default=None,
    required=False,
    type=click.INT,
    show_default=False,
    help=(
        "The minimum length (bases) of a read (shorter reads will be discarded).\n"
        "If you set this argument it will overrule the value from the chosen design"
    ),
)
@click.option(
    "--max-n-bases",
    default=0,
    required=False,
    type=click.INT,
    show_default=True,
    help=(
        "The maximum number of Ns allowed in a read (default of 0 means "
        "any reads with N in it will be filtered out)"
    ),
)
@click.option(
    "--avg-qual",
    default=20,
    required=False,
    type=click.INT,
    show_default=True,
    help="Minimum avg. quality a read must have (0 will disable the filter)",
)
@click.option(
    "--dedup",
    required=False,
    is_flag=True,
    help="Remove duplicated reads (exact same sequence)",
)
@click.option(
    "--remove-polyg",
    required=False,
    is_flag=True,
    help="Remove PolyG sequences (length of 10 or more)",
)
@output_option
@design_option
@click.pass_context
@timer
def preqc(
    ctx,
    fastq_file,
    trim_front,
    trim_tail,
    max_length,
    min_length,
    max_n_bases,
    avg_qual,
    dedup,
    remove_polyg,
    output,
    design,
):
    """Process raw Molecular Pixelation data (FASTQ) to perform QC, filtering,
    trimming and removal of duplicates.
    """  # noqa: D205,D400,D415
    # log input parameters
    input_files = [fastq_file]

    log_step_start(
        "fastp",
        input_files=input_files,
        output=output,
        trim_front=trim_front,
        trim_tail=trim_tail,
        max_length=max_length,
        min_length=min_length,
        max_n_bases=max_n_bases,
        avg_qual=avg_qual,
        dedup=dedup,
        remove_polyg=remove_polyg,
        design=design,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(
        input_files=input_files, allowed_extensions=("fastq.gz", "fq.gz")
    )

    # check that fastp is installed
    if which("fastp") is None:
        click.ClickException("Fastp is not installed in this system")

    # update arguments from config file (CLI arguments have priority)
    try:
        amplicon = config.get_assay(design).get_region_by_id("amplicon")
        conf_min_length, conf_max_length = amplicon.get_len()
        min_length = conf_min_length if min_length is None else min_length
        max_length = conf_max_length if max_length is None else max_length
    except KeyError as exc:
        raise click.ClickException(f"Parsing attribute from config file {exc}")

    # create output folder if it does not exist
    preqc_output = create_output_stage_dir(output, "preqc")

    # run fastq (pre QC and filtering) using parallel processing
    logger.info(f"Processing {fastq_file} with fastp")

    clean_name = get_sample_name(fastq_file)
    extension = get_extension(fastq_file)
    output_file = preqc_output / f"{clean_name}.processed.{extension}"
    failed_file = preqc_output / f"{clean_name}.failed.{extension}"
    html_file = preqc_output / f"{clean_name}.qc-report.html"
    json_file = preqc_output / f"{clean_name}.report.json"

    write_parameters_file(
        ctx,
        preqc_output / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-mpx preqc",
    )

    qc_fastq(
        input=str(fastq_file),
        output=str(output_file),
        failed=str(failed_file),
        report=str(html_file),
        metrics=str(json_file),
        design=design,
        n_limit=max_n_bases,
        trim_front=trim_front,
        trim_tail=trim_tail,
        min_length=min_length,
        max_length=max_length,
        # fastp developers recommend to keep this low to avoid I/O overhead
        threads=2,
        avg_qual=avg_qual,
        dedup=dedup,
        remove_polyg=remove_polyg,
        verbose=ctx.obj["VERBOSE"],
    )



"""Module for combining intermediate collapsed results into a final edgelist.

These routines are for combining data using the "independent" demuxing and collapsing strategy.

Copyright © 2025 Pixelgen Technologies AB
"""

import dataclasses
from collections import Counter
from pathlib import Path
from typing import Iterable, MutableMapping

import duckdb as dd
from duckdb import DuckDBPyConnection

from pixelator.pna.collapse.independent import (
    IndependentCollapseSampleReport,
    MarkerCorrectionStats,
    SingleUMICollapseSampleReport,
)
from pixelator.pna.collapse.paired.combine_collapse import logger
from pixelator.pna.utils import init_duckdb_conn


def _merge_sort_parquet(
    conn: DuckDBPyConnection, parquet: Iterable[Path], output_file: Path
) -> Path:
    """Combine and sort a partitioned collapsed dataset.

    DuckDB sorting can spill out to disk when memory space is limited.

    Args:
        conn: A DuckDB connection
        parquet: An iterable of input file paths
        output_file: The path to write the combined and sorted output to

    Returns:
        The path to the combined and sorted output file.

    """
    conn.sql(
        f"""
        CREATE OR REPLACE TEMP VIEW combined_sorted_data AS
        SELECT *
        FROM read_parquet({str([str(i) for i in parquet])})
        ORDER BY original_umi1, original_umi2, uei;
        """
    )
    conn.execute(
        f"""
        COPY combined_sorted_data TO '{output_file}' ( FORMAT 'parquet', CODEC 'zstd', COMPRESSION_LEVEL 6 );
        """
    )

    return output_file


@dataclasses.dataclass()
class CombineCollapseIndependentStats:
    """Some combine collapse statistics.

    Attributes:
        output_molecules: Total number of unique output molecules
        corrected_reads: Total number of corrected reads

    """

    output_molecules: int
    corrected_reads: int


def combine_independent_parquet_files(
    umi1_files: Iterable[Path],
    umi2_files: Iterable[Path],
    output_file: Path,
    threads: int | None = None,
    memory_limit: int | None = None,
    temp_directory: str | Path | None = None,
) -> CombineCollapseIndependentStats:
    """Scan a directory for parquet files with corrected UMI1s and UMI2s and join them.

    Use DuckDB for their support of larger than memory joins.

    The independently corrected partitioned umi1 and umi2 output files are first combined and sorted.
    This result in two sorted parquet files, one for with UMI1 correction info and one with the UMI2 correction.
    The sorting is done on the original_umi1,original_umi2 and uei columns.
    The umi regions before correction are common in the m1 and m2 parquet files,
    so we guaranteed that the row id of each molecule in the merged m1 file will match that
    of the row id in the m2 file.

    Combining the correction for the UMI1 and UMI2 can now be done using a POSITIONAL JOIN.

    The sorting and merging approach works well with DuckDB larger-than-memory processing.
    Simply merging and then using an INNER JOIN used a prohibitive amount of memory for our data
    which would try to join two tables with each in 100M+ rows.

    Note that we are also splitting the original molecule embedding into three separate integers
    (umi1, umi2, uei) in the demux step to optimize for the sorting operations here.

    This multi-key sorting helps with cache-locality in the sorting process and the use of integers
    enables DuckDB to use an efficient radix sort implementation.
    Sorting on the full molecule turned out to be too slow.

    See this conference paper for more details on DuckDB's sorting engine:
    https://hannes.muehleisen.org/publications/ICDE2023-sorting.pdf

    Args:
        umi1_files: The list of UMI1 parquet files.
        umi2_files: The list of UMI2 parquet files.
        output_file: The path to the output parquet file.
        threads: The number of threads to use for DuckDB. If None, the default is used (all available cores).
        memory_limit: The memory limit to use for DuckDB in MB eg: '1024MiB' for 1GiB.
            If None, the default is used (80% of the system memory).
        temp_directory: The directory used by DuckDB for disk spilling
        max_temp_directory_size:
            The maximum size in bytes that DuckDB is allowed to use for temporary files.
            If None, the default is used (90% of the system disk space).

        verbose: If True, enable DuckDB stdout logging.


    Returns:
        A dictionary with additional statistics calculated on the combined parquet data.

    """
    conn = init_duckdb_conn(
        memory_limit=memory_limit, threads=threads, temp_dir=temp_directory
    )

    logger.info("Combining and sorting UMI1 parquet files")
    sorted_umi1_file = _merge_sort_parquet(
        conn, umi1_files, output_file.with_suffix(".m1.parquet")
    )

    logger.info("Combining and sorting UMI2 parquet files")
    sorted_umi2_file = _merge_sort_parquet(
        conn, umi2_files, output_file.with_suffix(".m2.parquet")
    )

    logger.info("Combining UMI1 and UMI2 data.")

    # UMI1 data
    conn.execute(
        f"""
        CREATE OR REPLACE TEMP VIEW umi1_data AS
            SELECT marker_1, marker_2, read_count,
                   corrected_umi1 as umi1,
                   original_umi1,
                   corrected_umi1,
                   corrected_umi1 != original_umi1 as umi1_is_corrected,
                   uei
            FROM read_parquet('{str(sorted_umi1_file)}');
        """
    )

    # UMI2 data
    conn.execute(
        f"""
        CREATE OR REPLACE TEMP VIEW umi2_data AS
            SELECT corrected_umi2 as umi2,
                   original_umi2,
                   corrected_umi2 != original_umi2 as umi2_is_corrected
            FROM read_parquet('{str(sorted_umi2_file)}');
        """
    )

    conn.execute(
        """
        CREATE OR REPLACE TEMP VIEW combined_data AS
            SELECT marker_1, marker_2, umi1, umi2,
                sum(read_count)::UINTEGER as read_count,
                count(DISTINCT uei)::USMALLINT as uei_count,
                ifnull(sum("read_count") FILTER ( umi1_data.umi1_is_corrected OR umi2_data.umi2_is_corrected ), 0)::UINTEGER as corrected_read_count
            FROM umi1_data POSITIONAL JOIN umi2_data
            GROUP BY marker_1, marker_2, umi1, umi2
        """
    )

    conn.execute(
        f"""
        COPY combined_data TO '{output_file}' ( FORMAT 'parquet', CODEC 'zstd', COMPRESSION_LEVEL 6 );
        """
    )

    # Another pass to collect some stats needed for the reports
    res = conn.sql(
        f"""
        SELECT sum(uei_count)::UINTEGER as output_molecules,
               sum(corrected_read_count)::UINTEGER as corrected_reads
        FROM read_parquet('{output_file}')
        """
    ).pl()

    output_molecules, corrected_reads = res.row(0)

    stats = CombineCollapseIndependentStats(
        output_molecules=int(output_molecules), corrected_reads=int(corrected_reads)
    )

    # Sorted UMI files are only intermediates and can be removed
    sorted_umi1_file.unlink()
    sorted_umi2_file.unlink()

    conn.close()

    return stats


def combine_independent_report_files(
    marker1_reports: list[Path],
    marker2_reports: list[Path],
    sample_id: str,
    stats,
    output_file,
) -> None:
    """Combine collapse reports per marker into a single combined report.

    Args:
        marker1_reports: The list of marker1 reports.
        marker2_reports: The list of marker2 reports.
        sample_id: The sample ID to use for the combined report.
        stats: Additional statistics calculated on the combined parquet data.
            These statistics cannot be derived from the individual reports.
        output_file: The output file to write the combined report to.

    Returns:
        None

    """
    regions_to_combine = frozenset(
        (
            "input_reads",
            "input_molecules",
            "input_unique_umis",
            "output_unique_umis",
            "corrected_unique_umis",
            "corrected_reads",
        )
    )

    umi1_markers: list[MarkerCorrectionStats] = []
    umi1_summary_stats: MutableMapping[str, int] = Counter()

    for f in marker1_reports:
        umi1_report = SingleUMICollapseSampleReport.from_json(f)
        umi1_markers.extend(umi1_report.markers)
        umi1_summary_stats.update(
            {
                k: v
                for k, v in umi1_report.model_dump().items()
                if k in regions_to_combine
            }
        )

    umi2_markers: list[MarkerCorrectionStats] = []
    umi2_summary_stats: MutableMapping[str, int] = Counter()

    for f in marker2_reports:
        umi2_report = SingleUMICollapseSampleReport.from_json(f)
        umi2_markers.extend(umi2_report.markers)
        umi2_summary_stats.update(
            {
                k: v
                for k, v in umi2_report.model_dump().items()
                if k in regions_to_combine
            }
        )

    report = IndependentCollapseSampleReport(
        product_id="single-cell-pna",
        sample_id=sample_id,
        markers=umi1_markers + umi2_markers,
        input_reads=umi1_summary_stats["input_reads"],
        input_molecules=umi1_summary_stats["input_molecules"],
        corrected_reads=stats.corrected_reads,
        output_molecules=stats.output_molecules,
    )

    report.write_json_file(output_file, indent=4)
    return



"""Module relating to pixelator feature flags.

A feature flags optionally turns on features in pixelator that are not
yet ready for production. This is useful for testing and development work.

To use a feature flag in the code you can use the following pattern:

```
from pixelator.feature_flags import MY_FEATURE_FLAG

if MY_FEATURE_FLAG:
    # Do something
    ...

```

The value of the feature flag is determined by an environment variable,
which should have the name `PIXELATOR_<NAME_OF_FLAG>`.

Copyright © 2024 Pixelgen Technologies AB.
"""

import os
from typing import Any


def _load_flags_from_environment(flag: str, default: Any | None = None) -> Any:
    """Load feature flag from environment variables.

    :param flag: The name of the feature flag.
    :default: The default value of the feature flag.
    """
    return os.environ.get(f"PIXELATOR_{flag}", default)


MY_FEATURE_FLAG = bool(_load_flags_from_environment("MY_FEATURE_FLAG", default=False))



"""Pipeline step to identify the antibody for a given barcode.

Copyright © 2024 Pixelgen Technologies AB.
"""

from collections import Counter
from typing import Any, cast

from cutadapt.steps import HasFilterStatistics, SingleEndStep
from dnaio import SequenceRecord

from pixelator.pna.config import PNAAntibodyPanel, PNAAssay, get_position_in_parent
from pixelator.pna.demux.correction import BKTree, build_bktree, build_exact_dict_lookup
from pixelator.pna.read_processing.statistics import HasCustomStatistics


class BarcodeIdentifierStatistics:
    """Helper class to collect statistics for the barcode identifier step.

    Each individual worker in a PipelineExecutor will have its own
    independent instance of this class. At the end of the execution,
    the statistics will be merged into a single instance and passed
    to the final statistics object.

    :ivar passed: the number of reads that passed the barcode identifier
    :ivar failed: the number of reads that failed the barcode identifier
    :ivar missing_pid1: the number of reads with a missing PID1
    :ivar missing_pid2: the number of reads with a missing PID2
    :ivar pid1_matches_distance_distribution:
        A counter object capturing the distribution of distances for PID1 matches.
    :ivar pid2_matches_distance_distribution:
        A counter object capturing the distribution of distances for PID2 matches.

    :ivar pid_pair_counter: A counter object capturing the number of reads per PID pair.
    """

    def __init__(self):
        """Initialize the BarcodeIdentifierStatistics object."""
        # Passsed counters
        self.exact = 0
        self.corrected = 0

        # Failed counters
        self.missing_pid1 = 0
        self.missing_pid2 = 0
        self.missing_pid1_pid2 = 0

        self.pid1_matches_distance_distribution = Counter({})
        self.pid2_matches_distance_distribution = Counter({})

        self.pid_pair_counter = Counter({})

    @property
    def passed(self) -> int:
        """Return the number of reads that passed the barcode identification."""
        return self.exact + self.corrected

    @property
    def failed(self) -> int:
        """Return the total number of reads that failed the barcode identification step."""
        return self.missing_pid1_pid2 + self.missing_pid1 + self.missing_pid2

    def __iadd__(self, other):
        """Merge statistics from another object into this one."""
        if isinstance(other, BarcodeIdentifierStatistics):
            self.exact += other.exact
            self.corrected += other.corrected
            self.missing_pid1 += other.missing_pid1
            self.missing_pid2 += other.missing_pid2
            self.missing_pid1_pid2 += other.missing_pid1_pid2
            self.pid1_matches_distance_distribution += (
                other.pid1_matches_distance_distribution
            )
            self.pid2_matches_distance_distribution += (
                other.pid2_matches_distance_distribution
            )
            self.pid_pair_counter += other.pid_pair_counter
            return self

        return NotImplemented

    def get_pid_group_counter(self) -> dict[tuple[str, str], int]:
        """Return the number of reads per PID pair."""
        return dict(self.pid_pair_counter)

    @property
    def input(self) -> int:
        """Return the total number of reads processed."""
        return self.passed + self.failed

    def collect(self) -> dict[str, Any]:
        """Return a dictionary with statistics."""
        pid_pairs = list(
            (marker1, marker2, count)
            for (marker1, marker2), count in self.pid_pair_counter.items()
        )

        return {
            "input_reads": self.input,
            "output_reads": self.passed,
            "failed_reads": self.failed,
            "output_corrected_reads": self.corrected,
            "output_exact_reads": self.exact,
            "invalid_pid1_reads": self.missing_pid1,
            "invalid_pid2_reads": self.missing_pid2,
            "invalid_pid1_pid2_reads": self.missing_pid1_pid2,
            "pid1_matches_distance_distribution": dict(
                self.pid1_matches_distance_distribution
            ),
            "pid2_matches_distance_distribution": dict(
                self.pid2_matches_distance_distribution
            ),
            "pid_group_sizes": pid_pairs,
        }

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<BarcodeIdentifierStatistics [input={self.input} passed={self.passed} failed={self.failed}]>"


class BarcodeIdentifier(SingleEndStep, HasFilterStatistics, HasCustomStatistics):
    """A pipeline filter that finds the nearest antibody for a given barcode."""

    def __init__(
        self, assay: PNAAssay, panel: PNAAntibodyPanel, mismatches: int = 1, writer=None
    ):
        """Initialize the BarcodeIdentifier object.

        :param assay: the assay design
        :param panel: the antibody panel
        :param mismatches: the maximum number of mismatches allowed when aligning the LBS sequences
        :param writer: a writer to save failed reads to
        """
        self.assay = assay
        self.panel = panel
        self.mismatches = mismatches
        self._writer = writer

        self._tree_seq_1: BKTree = build_bktree(panel, sequence_key="sequence_1")
        self._exact_lookup_seq_1 = build_exact_dict_lookup(
            panel, sequence_key="sequence_1"
        )

        self._tree_seq_2: BKTree = build_bktree(panel, sequence_key="sequence_2")
        self._exact_lookup_seq_2 = build_exact_dict_lookup(
            panel, sequence_key="sequence_2"
        )

        self._pid1_slice = slice(*get_position_in_parent(assay, "pid-1"))
        self._pid2_slice = slice(*get_position_in_parent(assay, "pid-2"))

        self._stats = BarcodeIdentifierStatistics()

    def filtered(self):
        """Return the number of reads that passed the barcode identifier."""
        return self._stats.passed

    def descriptive_identifier(self):
        """Return a short name describing the step for the "filtered" section of the Statistics."""
        return "demux"

    def __call__(self, read: SequenceRecord, info=None) -> SequenceRecord | None:
        """Find the nearest antibody for a given barcode.

        :param read: the read to process
        :return: the read with the antibody information added
        """
        # Extract the barcode sequence
        _stats = self._stats

        pid1 = read.sequence[self._pid1_slice].encode("ascii")
        pid2 = read.sequence[self._pid2_slice].encode("ascii")

        # Check if we have an exact match first
        id1 = self._exact_lookup_seq_1.get(pid1)
        id2 = self._exact_lookup_seq_2.get(pid2)

        # Fast path - two exact matches
        if id1 and id2:
            _stats.pid1_matches_distance_distribution[0] += 1
            _stats.pid2_matches_distance_distribution[0] += 1
            _stats.exact += 1
            t = (id1, id2)
            _stats.pid_pair_counter[t] += 1
            read.name += f" {id1}:{id2}"
            return read

        # Error correcting path
        missing_pid1 = False
        missing_pid2 = False
        failed = False
        if not id1:
            nearest_matches1 = self._tree_seq_1.find(pid1, self.mismatches)
            if len(nearest_matches1) == 0:
                missing_pid1 = True
                failed = True
            else:
                pid1_score, pid1_node = nearest_matches1[0]
                _stats.pid1_matches_distance_distribution[pid1_score] += 1
                id1 = pid1_node.id

        if not id2:
            nearest_matches2 = self._tree_seq_2.find(pid2, self.mismatches)
            if len(nearest_matches2) == 0:
                missing_pid2 = True
                failed = True
            else:
                pid2_score, pid2_node = nearest_matches2[0]
                _stats.pid1_matches_distance_distribution[pid2_score] += 1
                id2 = pid2_node.id

        if failed:
            if missing_pid1 and missing_pid2:
                _stats.missing_pid1_pid2 += 1
            elif missing_pid1:
                _stats.missing_pid1 += 1
            elif missing_pid2:
                _stats.missing_pid2 += 1

            read.name += f" {id1}:{id2}"
            if self._writer:
                self._writer.write(read)
            return None

        # Mypy cannot infer that id1 and id2 are not None here
        t = cast(tuple[str, str], (id1, id2))
        _stats.pid_pair_counter[t] += 1
        _stats.corrected += 1

        read.name += f" {id1}:{id2}"
        return read

    def get_statistics_name(self):
        """Return a short name describing the statistics object."""
        return "demux"

    def get_statistics(self) -> BarcodeIdentifierStatistics:
        """Return the statistics object."""
        return self._stats



"""Graph console script for pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.cli.common import logger, output_option
from pixelator.mpx.graph import connect_components


@click.command(
    "graph",
    short_help=("compute graph, components and metrics from an edge list"),
    options_metavar="<options>",
)
@click.argument(
    "parquet_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="parquet",
)
@click.option(
    "--multiplet-recovery",
    required=False,
    is_flag=True,
    default=False,
    type=click.BOOL,
    help=("Activate the multiplet recovery using leiden community detection"),
)
@click.option(
    "--max-refinement-recursion-depth",
    default=5,
    required=False,
    type=click.IntRange(1, 100),
    show_default=True,
    help=(
        "The number of times a component can be broken down into "
        "smaller components during the multiplet recovery process."
    ),
)
@click.option(
    "--max-edges-to-split",
    default=5,
    required=False,
    type=click.IntRange(1, 100),
    show_default=True,
    help=(
        "Maximum number of edges  between the product components "
        "as a result of a component split operation during "
        "the multiplet recovery process."
    ),
)
@click.option(
    "--min-count",
    default=2,
    required=False,
    type=click.IntRange(1, 50),
    show_default=True,
    help=(
        "Discard edges (pixels) with with a count (reads) below this (use 1 to disable)"
    ),
)
@output_option
@click.pass_context
@timer
def graph(
    ctx,
    parquet_file,
    multiplet_recovery,
    max_refinement_recursion_depth,
    max_edges_to_split,
    min_count,
    output,
):
    """Compute graph, components and other metrics from an edge list."""
    # log input parameters
    input_files = [parquet_file]
    log_step_start(
        "graph",
        input_files=input_files,
        output=output,
        multiplet_recovery=multiplet_recovery,
        max_refinement_recursion_depth=max_refinement_recursion_depth,
        max_edges_to_split=max_edges_to_split,
        min_count=min_count,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files, allowed_extensions="parquet")

    # create output folder if it does not exist
    graph_output = create_output_stage_dir(output, "graph")

    # compute graph/components using parallel processing
    logger.info(f"Computing clusters for file {parquet_file}")

    clean_name = get_sample_name(parquet_file)
    metrics_file = graph_output / f"{clean_name}.report.json"

    write_parameters_file(
        ctx,
        graph_output / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-mpx graph",
    )

    connect_components(
        input=str(parquet_file),
        output=str(graph_output),
        sample_name=clean_name,
        metrics_file=str(metrics_file),
        multiplet_recovery=multiplet_recovery,
        max_refinement_recursion_depth=max_refinement_recursion_depth,
        max_edges_to_split=max_edges_to_split,
        min_count=min_count,
    )



"""Utility functions for parsing numbers with optional unit suffixes.

Copyright © 2024 Pixelgen Technologies AB.
"""

import re

units = {"K": 10**3, "M": 10**6, "G": 10**9}


def parse_size(s: str) -> int | float:
    """Parse a string as a number with optional unit suffix [K, M, G].

    :param s: The string to parse
    :return: The parsed number as a float
    """
    match = re.match(r"(?P<value>\d+(?:.\d+)?)(?P<unit>[KMGkmg])?$", s)
    if not match:
        raise ValueError(f"Invalid number: {s}")

    number = float(match.group("value"))
    unit = match.group("unit")

    unit_scale = 1
    if unit:
        unit_scale = units.get(unit.upper())  # type: ignore
        if unit_scale is None:
            raise ValueError(f"Invalid unit: {unit}")

    result = float(number) * unit_scale
    if result.is_integer():
        return int(result)
    return result



"""Pipeline for processing amplicon reads, code adopted from cutadapt.

Modifications Copyright © 2025 Pixelgen Technologies AB.
Under the same license terms as the original code.

Original copyright notice:

Copyright (c) 2010 Marcel Martin <marcel.martin@scilifelab.se> and contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
import typing
from typing import Iterable, Optional, Tuple

from cutadapt.files import InputFiles
from cutadapt.info import ModificationInfo
from cutadapt.modifiers import (
    PairedEndModifier,
    PairedEndModifierWrapper,
    SingleEndModifier,
)
from cutadapt.pipeline import Pipeline
from cutadapt.steps import PairedEndStep, SingleEndStep
from cutadapt.utils import Progress

from pixelator.pna.read_processing.modifiers import CombiningModifier

PairedModifiers = (
    PairedEndModifier | tuple[SingleEndModifier | None, SingleEndModifier | None]
)


ModifierStage = typing.Literal["pre", "post"]


class AmpliconPipeline(Pipeline):
    """Processing pipeline for processing single-end or paired‐end reads into a single amplicon sequence.

    - If two files (R1+R2) are provided: pre‐steps (paired), then combine, then post‐steps (single).
    - If only one file is provided (single‐end), you detect “R1” vs “R2” by looking at
      read.name, and treat that one FASTQ as if it were already “combined.”  All
      single‐end modifiers/steps run in that branch.

    :param combiner: The step that combines the reads into a single amplicon sequence.
    :param pre_modifiers: A list of modifiers that are applied to the reads before the combining step.
    :param pre_steps: A list of steps that are applied to the reads before the combining step.
    :param post_modifiers: A list of modifiers that are applied to the reads after the combining step.
    :param post_steps: A list of steps that are applied to the reads after the combining step.
    """

    paired: bool | None = None

    def __init__(
        self,
        combiner: CombiningModifier,
        pre_modifiers: (
            Iterable[
                PairedEndModifier
                | SingleEndModifier
                | tuple[
                    SingleEndModifier | None,
                    SingleEndModifier | None,
                ]
            ]
            | None
        ) = None,
        pre_steps: Iterable[PairedEndStep | SingleEndStep] | None = None,
        post_modifiers: Iterable[SingleEndModifier] | None = None,
        post_steps: Iterable[SingleEndStep] | None = None,
    ):
        """Initialize the pipeline.

        :param combiner: The step that combines the reads into a single
        amplicon sequence. In single-end mode, this step works as a modifier
        to enforce the expected read format.
        :param pre_modifiers: A list of modifiers that are applied to the reads before the combining step.
        :param pre_steps: A list of steps that are applied to the reads before the combining step.
        :param post_modifiers: A list of modifiers that are applied to the reads after the combining step.
        :param post_steps: A list of steps that are applied to the reads after the combining step.
        """
        self._combiner = combiner
        self._pre_modifiers: list[
            SingleEndModifier
            | PairedEndModifier
            | tuple[SingleEndModifier | None, SingleEndModifier | None]
        ] = []
        self._pre_steps: list[PairedEndStep | SingleEndStep] = (
            list(pre_steps) if pre_steps else []
        )
        self._post_modifiers: list[SingleEndModifier] = []
        self._post_steps: list[SingleEndStep] = list(post_steps) if post_steps else []

        self._reader = None

        # Whether to ignore pair_filter mode for discard-untrimmed filter
        self.override_untrimmed_pair_filter = False

        # Load the pre- and post-modifiers from the constructor arguments
        if pre_modifiers:
            self._add_modifiers(pre_modifiers, stage="pre")
        if post_modifiers:
            self._add_modifiers(post_modifiers, stage="post")

    @property
    def _modifiers(self) -> list[PairedEndModifier]:
        """Return a list of all modifiers in the pipeline."""
        return self._pre_modifiers + self._post_modifiers

    @property
    def _steps(self) -> list[PairedEndModifier]:
        """Return a list of all steps in the pipeline."""
        return self._pre_steps + [self._combiner] + self._post_steps

    def _add_modifiers(
        self,
        modifiers: Iterable[
            SingleEndModifier
            | PairedEndModifier
            | Tuple[Optional[SingleEndModifier], Optional[SingleEndModifier]],
        ],
        stage: ModifierStage,
    ) -> None:
        for modifier in modifiers:
            if isinstance(modifier, tuple):
                self._add_two_single_modifiers(modifier[0], modifier[1], stage=stage)
            else:
                self._add_modifier(modifier, stage=stage)

    def _add_two_single_modifiers(
        self,
        modifier1: Optional[SingleEndModifier],
        modifier2: Optional[SingleEndModifier],
        stage: ModifierStage,
    ) -> None:
        """Add two single-end modifiers that modify R1 and R2, respectively.

        One of them can be None, in which case the modifier
        is only applied to the respective other read.

        :param modifier1: The modifier for read 1.
        :param modifier2: The modifier for read 2.
        :param stage: The stage in which to apply the modifiers
        """
        assert stage in ("pre", "post")

        if modifier1 is None and modifier2 is None:
            raise ValueError("Not both modifiers can be None")

        if stage == "pre":
            self._pre_modifiers.append(PairedEndModifierWrapper(modifier1, modifier2))
        if stage == "post":
            self._post_modifiers.append(PairedEndModifierWrapper(modifier1, modifier2))

    def _add_modifier(self, modifier: PairedEndModifier, stage: ModifierStage) -> None:
        """Add a Modifier (without wrapping it in a PairedEndModifierWrapper)."""
        assert stage in ("pre", "post")

        if stage == "pre":
            self._pre_modifiers.append(modifier)
        if stage == "post":
            self._post_modifiers.append(modifier)

    def _pre_process_paired(self, reads):
        """Pre‐process paired reads."""
        pre_modifiers_and_steps = self._pre_modifiers + self._pre_steps
        read1, read2 = reads
        n_bp1 = len(read1)
        n_bp2 = len(read2)
        info1 = ModificationInfo(read1)
        info2 = ModificationInfo(read2)

        for step in pre_modifiers_and_steps:
            reads = step(*reads, info1, info2)  # type: ignore
            if reads is None:
                break

        if reads is not None:
            read1, read2 = reads

        return read1, read2, info1, info2, n_bp1, n_bp2

    def _pre_process_single(self, single_read):
        """Pre‐process single reads."""
        pre_modifiers_and_steps = self._pre_modifiers + self._pre_steps
        info = ModificationInfo(single_read)

        for modifier in pre_modifiers_and_steps:
            out = modifier(single_read, info)
            if out is None:
                break

        return out, info, len(single_read)

    def process_reads(
        self,
        infiles: InputFiles,
        progress: Optional[Progress] = None,
    ) -> Tuple[int, int, Optional[int]]:
        """Receive a slice of reads and process them through the pipeline.

        :param infiles: A list of input files to process.
        :returns: (n_reads, total1_bp, total2_bp or None).
        """
        self._infiles = infiles
        self._reader = infiles.open()

        if len(infiles._files) == 1:
            self.paired = False
        elif len(infiles._files) == 2:
            self.paired = True
        else:
            raise ValueError(
                "AmpliconPipeline requires either one or two input files (single‐end or paired‐end)."
            )

        n = 0  # no. of processed reads
        total1_bp = 0
        total2_bp = 0 if self.paired else None
        assert self._reader is not None

        post_modifiers_and_steps = self._post_modifiers + self._post_steps

        # Note that this reader is not a real file but a chunk backed by a BytesIO object
        # in case the pipeline is running on the ParallelPipelineRunner.
        # Each worker will have its own chunk of reads to process in its own pipeline instance.
        for reads in self._reader:
            n += 1
            if (n % 10000 == 0) and (progress is not None):
                progress.update(10000)
            if self.paired:
                read1, read2, info1, info2, n_bp1, n_bp2 = self._pre_process_paired(
                    reads
                )
                total1_bp += n_bp1
                total2_bp += n_bp2
                final_read = self._combiner(read1, read2, info1, info2)

            else:
                read, info, n_bp = self._pre_process_single(reads)
                total1_bp += n_bp
                # Figure out which read to pass as None, since the combiner expects two reads.
                if read.name.endswith("/2") or read.name.endswith("_R2"):
                    final_read = self._combiner(None, read, None, info)
                else:
                    final_read = self._combiner(read, None, info, None)

            final_read_info = ModificationInfo(final_read)

            if final_read is not None:
                for step in post_modifiers_and_steps:
                    try:
                        read = step(final_read, final_read_info)  # type: ignore
                    except Exception as e:
                        logging.error(
                            "Error in step %s for read %s", step, final_read.name
                        )
                        raise

                    if read is None:
                        break

        if progress is not None:
            progress.update(n % 10000)

        infiles.close()
        return (n, total1_bp, total2_bp)



"""Model for flow of input/output of read counts through processing stages.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import typing

import pydantic

from pixelator.pna.report.models.base import SampleReport


class ReadsDataflowReport(SampleReport):
    """Model for tracking the read counts through all processing stages."""

    report_type: typing.Literal["reads_dataflow"] = "reads_dataflow"

    input_read_count: int = pydantic.Field(
        ...,
        description="The number of raw input reads from the input fastq files.",
    )

    amplicon_output_read_count: int = pydantic.Field(
        ...,
        description="The number of valid reads after QC filtering and amplicon assembly.",
    )

    demux_output_read_count: int = pydantic.Field(
        ...,
        description="The number of input reads after QC filtering that could be demultiplexed.",
    )

    collapse_output_molecule_count: int = pydantic.Field(
        ...,
        description="The number of reads remaining after error-correction to known antibody markers.",
    )

    graph_output_molecule_count: int = pydantic.Field(
        ...,
        description="The number of molecules in the proximity graph.",
    )



"""Custom JSON encoder for use in Pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

import dataclasses
import json
from typing import Any

from pydantic import BaseModel


class PixelatorJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder for use in Pixelator.

    Support pydantic models and dataclasses serialization.
    """

    def default(self, obj: Any) -> Any:
        """Return a serializable object for ``obj``.

        Dump pydantic models to json or calls the base class implementation
        for other types.

        :param obj: object to serialize
        :returns: a serializable object
        """
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json")
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)  # type: ignore
        return super().default(obj)
